import array
from pathlib import Path

import torch

from ...bi_encoder import BiEncoderModule, BiEncoderOutput
from ...data import IndexBatch
from ...models import ColConfig, DprConfig
from ..base import IndexConfig, Indexer


class TorchDenseIndexer(Indexer):
    """Indexer for dense embeddings using PyTorch."""

    def __init__(
        self,
        index_dir: Path,
        index_config: "TorchDenseIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the TorchDenseIndexer.

        Args:
            index_dir (Path): Directory to store the index.
            index_config (TorchDenseIndexConfig): Configuration for the dense index.
            module (BiEncoderModule): Bi-encoder module to use for indexing.
            verbose (bool): Whether to print verbose output. Defaults to False.
        """
        super().__init__(index_dir, index_config, module, verbose)
        self.embeddings = array.array("f")

    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None:
        """Add embeddings from the output to the index.

        Args:
            index_batch (IndexBatch): Batch containing the index data.
            output (BiEncoderOutput): Output from the Bi-encoder model containing embeddings.
        Raises:
            ValueError: If output does not contain document embeddings.
        """
        doc_embeddings = output.doc_embeddings
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")

        if doc_embeddings.scoring_mask is None:
            doc_lengths = torch.ones(
                doc_embeddings.embeddings.shape[0], device=doc_embeddings.device, dtype=torch.int32
            )
            embeddings = doc_embeddings.embeddings[:, 0]
        else:
            doc_lengths = doc_embeddings.scoring_mask.sum(dim=1)
            embeddings = doc_embeddings.embeddings[doc_embeddings.scoring_mask]
        num_docs = len(index_batch.doc_ids)
        self.doc_ids.extend(index_batch.doc_ids)
        self.doc_lengths.extend(doc_lengths.int().cpu().tolist())
        self.num_embeddings += embeddings.shape[0]
        self.num_docs += num_docs
        self.embeddings.extend(embeddings.cpu().view(-1).float().tolist())

    def to_gpu(self) -> None:
        """Convert the index to GPU format."""
        pass

    def to_cpu(self) -> None:
        """Convert the index to CPU format."""
        pass

    def save(self) -> None:
        """Save the index to the specified directory."""
        super().save()
        index = torch.frombuffer(self.embeddings, dtype=torch.float32).view(self.num_embeddings, -1)
        torch.save(index, self.index_dir / "index.pt")


class TorchDenseIndexConfig(IndexConfig):
    """Configuration for the TorchDenseIndexer."""

    indexer_class = TorchDenseIndexer
    SUPPORTED_MODELS = {ColConfig.model_type, DprConfig.model_type}



"""
Base configuration class for Lightning IR models.

This module defines the configuration class `LightningIRConfig` which is used to instantiate
a Lightning IR model. The configuration class acts as a mixin for the `transformers.PretrainedConfig`
class from the Hugging Face Transformers library.
"""

from __future__ import annotations

import inspect
from pathlib import Path
from typing import TYPE_CHECKING, Any

from transformers import PretrainedConfig

from .class_factory import LightningIRConfigClassFactory
from .external_model_hub import CHECKPOINT_MAPPING

if TYPE_CHECKING:
    from .tokenizer import LightningIRTokenizer

    try:
        from peft import LoraConfig
    except ImportError:

        class LoraConfig:
            pass


class LightningIRConfig(PretrainedConfig):
    """The configuration class to instantiate a Lightning IR model. Acts as a mixin for the
    transformers.PretrainedConfig_ class.

    .. _transformers.PretrainedConfig: \
https://huggingface.co/transformers/main_classes/configuration.html#transformers.PretrainedConfig
    """

    model_type = "lightning-ir"
    """Model type for the configuration."""
    backbone_model_type: str | None = None
    """Backbone model type for the configuration. set by :func:`LightningIRModelClassFactory`."""

    def __init__(
        self,
        *args,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        use_adapter: bool = False,
        adapter_config: LoraConfig | None = None,
        pretrained_adapter_name_or_path: str | None = None,
        **kwargs,
    ):
        """Initializes the configuration.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            use_adapter (bool, optional): Whether to use LoRA adapters. Defaults to False.
            adapter_config (Optional[LoraConfig], optional): Configuration for LoRA adapters.
                Only used if use_adapter is True. Defaults to None.
            pretrained_adapter_name_or_path (Optional[str], optional): The path to a pretrained adapter to load.
                Defaults to None.
        """
        super().__init__(*args, **kwargs)
        self.query_length = query_length
        self.doc_length = doc_length
        self.use_adapter = use_adapter
        self.adapter_config = adapter_config
        self.pretrained_adapter_name_or_path = pretrained_adapter_name_or_path

    def get_tokenizer_kwargs(self, Tokenizer: type[LightningIRTokenizer]) -> dict[str, Any]:
        """Returns the keyword arguments for the tokenizer. This method is used to pass the configuration
        parameters to the tokenizer.

        Args:
            Tokenizer (type[LightningIRTokenizer]): Class of the tokenizer to be used.
        Returns:
            dict[str, Any]: Keyword arguments for the tokenizer.
        """
        return {k: getattr(self, k) for k in inspect.signature(Tokenizer.__init__).parameters if hasattr(self, k)}

    def to_dict(self) -> dict[str, Any]:
        """Overrides the transformers.PretrainedConfig.to_dict_ method to include the added arguments and the backbone
        model type.

        .. _transformers.PretrainedConfig.to_dict: \
https://huggingface.co/docs/transformers/en/main_classes/configuration#transformers.PretrainedConfig.to_dict

        Returns:
            dict[str, Any]: Configuration dictionary.
        """
        output = super().to_dict()
        if self.backbone_model_type is not None:
            output["backbone_model_type"] = self.backbone_model_type
        return output

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str | Path, *args, **kwargs) -> LightningIRConfig:
        """Loads the configuration from a pretrained model. Wraps the transformers.PretrainedConfig.from_pretrained_

        .. _transformers.PretrainedConfig.from_pretrained: \
https://huggingface.co/docs/transformers/en/main_classes/configuration#transformers.PretrainedConfig.from_pretrained

        Args:
            pretrained_model_name_or_path (str | Path): Pretrained model name or path.
        Returns:
            LightningIRConfig: Derived LightningIRConfig class.
        Raises:
            ValueError: If `pretrained_model_name_or_path` is not a Lightning IR model and no
                :py:class:`LightningIRConfig` is passed.
        """
        # provides AutoConfig.from_pretrained support
        if cls is LightningIRConfig or all(issubclass(base, LightningIRConfig) for base in cls.__bases__):
            # no backbone config found, create derived lightning-ir config based on backbone config
            config = None
            if pretrained_model_name_or_path in CHECKPOINT_MAPPING:
                config = CHECKPOINT_MAPPING[pretrained_model_name_or_path]
                ConfigClass = config.__class__
            elif cls is not LightningIRConfig:
                ConfigClass = cls
            else:
                ConfigClass = type(LightningIRConfigClassFactory.get_lightning_ir_config(pretrained_model_name_or_path))
                if ConfigClass is None:
                    raise ValueError("Pass a config to `from_pretrained`.")
            backbone_config = LightningIRConfigClassFactory.get_backbone_config(pretrained_model_name_or_path)
            cls = LightningIRConfigClassFactory(ConfigClass).from_backbone_class(type(backbone_config))
            if config is not None and all(issubclass(base, LightningIRConfig) for base in config.__class__.__bases__):
                derived_config = cls.from_pretrained(pretrained_model_name_or_path, config=config)
                derived_config.update(config.to_dict())
            return cls.from_pretrained(pretrained_model_name_or_path, *args, **kwargs)
        return super().from_pretrained(pretrained_model_name_or_path, *args, **kwargs)



import pytest

from lightning_ir import RunDataset


@pytest.mark.dataset
@pytest.mark.parametrize(
    "ir_datasets_run_path_or_id",
    [
        "msmarco-passage/train/colbert-docpairs",
        "msmarco-passage/train/kd-docpairs",
        "msmarco-passage/train/rank-distillm-colbert",
        "msmarco-passage/train/rank-distillm-monoelectra",
        "msmarco-passage/train/rank-distillm-rankzephyr",
        "msmarco-passage/train/rank-distillm-set-encoder",
        "msmarco-passage/train/sbert-bm25-docpairs",
        "msmarco-passage/train/sbert-msmarco-distilbert-base-tas-b-docpairs",
        "msmarco-passage/train/sbert-msmarco-distilbert-base-v3-docpairs",
        "msmarco-passage/train/sbert-msmarco-MiniLM-L-6-v3-docpairs",
    ],
)
def test_external_dataset(ir_datasets_run_path_or_id: str):
    dataset = RunDataset(ir_datasets_run_path_or_id).prepare_data()
    assert dataset is not None



"""
Module for indexing and searching using PLAID.

This module provides classes and configurations for indexing and searching using PLAID in the Lightning IR framework.

PLAID: `PLAID: An Efficient Engine for Late Interaction Retrieval
<https://dl.acm.org/doi/10.1145/3511808.3557325>`_
"""

from .plaid_indexer import PlaidIndexConfig, PlaidIndexer
from .plaid_searcher import PlaidSearchConfig, PlaidSearcher

__all__ = ["PlaidIndexConfig", "PlaidIndexer", "PlaidSearchConfig", "PlaidSearcher"]



"""
Adapter module for Lightning IR models.

This module provides LoRA adapter support for Lightning IR models using the PEFT library.
The adapter functionality is optional and only enabled when explicitly configured.
"""

from __future__ import annotations

try:
    from peft import LoraConfig, get_peft_model

    PEFT_AVAILABLE = True
except ImportError:
    PEFT_AVAILABLE = False


class LightningIRAdapterMixin:
    """Mixin class that adds LoRA adapter functionality to Lightning IR models."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._adapter_enabled = False
        self._hf_peft_config_loaded = False

    def init_adapters(self, adapter_config: LoraConfig) -> None:
        """Enable LoRA adapters on the model.

        Args:
            adapter_config: Configuration for the LoRA adapter.

        Raises:
            ImportError: If PEFT is not available.
            ValueError: If adapters are already enabled.
        """
        if not PEFT_AVAILABLE:
            raise ImportError(
                "PEFT is required for adapter functionality. Install it with: pip install lightning-ir[adapters]"
            )

        if self._hf_peft_config_loaded:
            raise ValueError("Adapters are already enabled on this model")

        peft_model = get_peft_model(self, adapter_config)

        for name, module in peft_model.named_children():
            if hasattr(self, name) and name != "base_model":
                original_module = getattr(self, name)
                if original_module is not module:  # Only set if it's actually different
                    setattr(self, name, module)

        self._adapter_enabled = True
        self._hf_peft_config_loaded = True

    def disable_adapters(self) -> None:
        """Disable LoRA adapters."""
        if not self._adapter_enabled:
            return
        if hasattr(self, "disable_adapter_layers"):
            self.disable_adapter_layers()
        elif hasattr(self, "disable_adapter"):
            self.disable_adapter()

    def enable_adapters(self) -> None:
        """(Re-)Enable LoRA adapters."""
        if self._adapter_enabled:
            return
        if hasattr(self, "enable_adapter_layers"):
            self.enable_adapter_layers()
        elif hasattr(self, "enable_adapter"):
            self.enable_adapter()



"""Module containing callbacks for indexing, searching, ranking, and registering custom datasets."""

from __future__ import annotations

import csv
import gc
import itertools
from dataclasses import is_dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, TypeVar

import pandas as pd
import torch
from lightning import Trainer
from lightning.pytorch.callbacks import Callback, TQDMProgressBar

from ..base.validation_utils import evaluate_run
from ..data import LightningIRDataModule, RankBatch, SearchBatch
from ..data.dataset import RUN_HEADER, DocDataset, IRDataset, QueryDataset, RunDataset, _DummyIterableDataset
from ..data.external_datasets.ir_datasets_utils import register_new_dataset
from ..retrieve import IndexConfig, Indexer, SearchConfig, Searcher

if TYPE_CHECKING:
    from ..base import LightningIRModule, LightningIROutput
    from ..bi_encoder import BiEncoderModule, BiEncoderOutput

T = TypeVar("T")


def _format_large_number(number: float) -> str:
    suffixes = ["", "K", "M", "B", "T"]
    suffix_index = 0

    while number >= 1000 and suffix_index < len(suffixes) - 1:
        number /= 1000.0
        suffix_index += 1

    formatted_number = f"{number:.2f}"

    suffix = suffixes[suffix_index]
    if suffix:
        formatted_number += f" {suffix}"
    return formatted_number


class _GatherMixin:
    """Mixin to gather dataclasses across all processes"""

    def _gather(self, pl_module: LightningIRModule, dataclass: T) -> T:
        if is_dataclass(dataclass):
            return dataclass.__class__(
                **{k: self._gather(pl_module, getattr(dataclass, k)) for k in dataclass.__dataclass_fields__}
            )
        return pl_module.all_gather(dataclass)


class _IndexDirMixin:
    """Mixin to get index_dir"""

    index_dir: Path | str | None
    index_name: str | None

    def _get_index_dir(self, pl_module: BiEncoderModule, dataset: DocDataset) -> Path:
        index_dir = self.index_dir
        if index_dir is None:
            default_index_dir = Path(pl_module.config.name_or_path)
            if default_index_dir.exists():
                index_dir = default_index_dir / "indexes"
            else:
                raise ValueError("No index_dir provided and model_name_or_path is not a path")
        index_dir = Path(index_dir)
        if self.index_name is None:
            index_dir = index_dir / dataset.dashed_docs_dataset_id
        else:
            index_dir = index_dir / self.index_name
        return index_dir


class _OverwriteMixin:
    """Mixin to skip datasets (for indexing or searching) if they already exist"""

    _get_save_path: Callable[[LightningIRModule, IRDataset], Path]

    def _remove_overwrite_datasets(self, trainer: Trainer, pl_module: LightningIRModule) -> None:
        overwrite = getattr(self, "overwrite", False)
        if not overwrite:
            datamodule: LightningIRDataModule | None = getattr(trainer, "datamodule", None)
            if datamodule is None:
                raise ValueError("No datamodule found")
            if datamodule.inference_datasets is None:
                return
            inference_datasets = list(datamodule.inference_datasets)
            for dataset in inference_datasets:
                save_path = self._get_save_path(pl_module, dataset)
                if save_path.exists():
                    dataset._SKIP = True
                    trainer.print(f"`{save_path}` already exists. set overwrite=True to overwrite")
                    if (
                        save_path.name.endswith(".run")
                        and dataset.qrels is not None
                        and pl_module.evaluation_metrics is not None
                    ):
                        run = RunDataset._load_csv(save_path)
                        qrels = dataset.qrels.stack(future_stack=True).dropna().astype(int).reset_index()
                        if isinstance(dataset, RunDataset) and dataset.run_path is not None:
                            dataset_id = dataset.run_path.name
                        else:
                            dataset_id = dataset.dataset_id
                        for key, value in evaluate_run(run, qrels, pl_module.evaluation_metrics).items():
                            key = f"{dataset_id}/{key}"
                            pl_module._additional_log_metrics[key] = value

    def _cleanup(self, trainer: Trainer, pl_module: LightningIRModule) -> None:
        # reset skip flat and additional log metrics
        datamodule: LightningIRDataModule | None = getattr(trainer, "datamodule", None)
        if datamodule is not None and datamodule.inference_datasets is not None:
            for dataset in datamodule.inference_datasets:
                dataset._SKIP = False
        pl_module._additional_log_metrics = {}


class IndexCallback(Callback, _GatherMixin, _IndexDirMixin, _OverwriteMixin):
    def __init__(
        self,
        index_config: IndexConfig,
        index_dir: Path | str | None = None,
        index_name: str | None = None,
        overwrite: bool = False,
        verbose: bool = False,
    ) -> None:
        """Callback to index documents using an :py:class:`~lightning_ir.retrieve.base.indexer.Indexer`.

        Args:
            index_config (IndexConfig): Configuration for the indexer.
            index_dir (Path | str | None): Directory to save index(es) to. If None, indexes will be stored in the
                model's directory. Defaults to None.
            index_name (str | None): Name of the index. If None, the dataset's dataset_id or file name will be used.
                Defaults to None.
            overwrite (bool): Whether to skip or overwrite already existing indexes. Defaults to False.
            verbose (bool): Toggle verbose output. Defaults to False.
        """
        super().__init__()
        self.index_config = index_config
        self.index_dir = index_dir
        self.index_name = index_name
        self.overwrite = overwrite
        self.verbose = verbose
        self.indexer: Indexer

    def setup(self, trainer: Trainer, pl_module: BiEncoderModule, stage: str) -> None:
        """Hook to setup the callback.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module used for indexing.
            stage (str): Stage of the trainer, must be "test".
        Raises:
            ValueError: If the stage is not "test".
        """
        if stage != "test":
            raise ValueError(f"{self.__class__.__name__} can only be used in test stage")
        self._remove_overwrite_datasets(trainer, pl_module)

    def teardown(self, trainer: Trainer, pl_module: BiEncoderModule, stage: str) -> None:
        """Hook to cleanup the callback.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module used for indexing.
            stage (str): Stage of the trainer.
        """
        self._cleanup(trainer, pl_module)

    def _get_save_path(self, pl_module: BiEncoderModule, dataset: IRDataset) -> Path:
        if not isinstance(dataset, DocDataset):
            raise ValueError("Expected DocDataset for indexing")
        return self._get_index_dir(pl_module, dataset)

    def _get_indexer(self, pl_module: BiEncoderModule, dataloader_idx: int) -> Indexer:
        dataset = pl_module.get_dataset(dataloader_idx)
        if dataset is None:
            raise ValueError("No dataset found to index")
        if not isinstance(dataset, DocDataset):
            raise ValueError("Expected DocDataset for indexing")
        index_dir = self._get_save_path(pl_module, dataset)

        indexer = self.index_config.indexer_class(index_dir, self.index_config, pl_module, self.verbose)
        return indexer

    def _log_to_pg(self, info: dict[str, Any], trainer: Trainer):
        pg_callback = trainer.progress_bar_callback
        if pg_callback is None or not isinstance(pg_callback, TQDMProgressBar):
            return
        pg = pg_callback.test_progress_bar
        info = {k: _format_large_number(v) for k, v in info.items()}
        if pg is not None:
            pg.set_postfix(info)

    def on_test_start(self, trainer: Trainer, pl_module: BiEncoderModule) -> None:
        """Hook to test datasets are configured correctly.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module.
        Raises:
            ValueError: If no test_dataloaders are found.
            ValueError: If not all test datasets are :py:class:`~lightning_ir.data.dataset.DocDataset`.
        """
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        datasets = [dataloader.dataset for dataloader in dataloaders]
        if not all(isinstance(dataset, (DocDataset, _DummyIterableDataset)) for dataset in datasets):
            raise ValueError("Expected DocDatasets for indexing")

    def on_test_batch_start(
        self, trainer: Trainer, pl_module: BiEncoderModule, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Hook to setup the indexer between datasets.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module.
            batch (Any): Batch of input data.
            batch_idx (int): Index of batch in the current dataset.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        """
        if batch_idx == 0:
            self.indexer = self._get_indexer(pl_module, dataloader_idx)
        super().on_test_batch_start(trainer, pl_module, batch, batch_idx, dataloader_idx)

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: BiEncoderModule,
        outputs: BiEncoderOutput,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Hook to pass encoded documents to the indexer

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module.
            outputs (BiEncoderOutput): Encoded documents.
            batch (Any): Batch of input data.
            batch_idx (int): Index of batch in the current dataset.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        """
        batch = self._gather(pl_module, batch)
        outputs = self._gather(pl_module, outputs)

        if not trainer.is_global_zero:
            return

        self.indexer.add(batch, outputs)
        self._log_to_pg(
            {
                "num_docs": self.indexer.num_docs,
                "num_embeddings": self.indexer.num_embeddings,
            },
            trainer,
        )
        # TODO if dataset length cannot be inferred, num_test_batches is inf and no index is saved
        if batch_idx == trainer.num_test_batches[dataloader_idx] - 1:
            assert hasattr(self, "indexer")
            self.indexer.save()


class RankCallback(Callback, _GatherMixin, _OverwriteMixin):
    def __init__(
        self, save_dir: Path | str | None = None, run_name: str | None = None, overwrite: bool = False
    ) -> None:
        """Callback to write run file of ranked documents to disk.

        Args:
            save_dir (Path | str | None): Directory to save run files to. If None, run files will be saved in the
                models' directory. Defaults to None.
            run_name (str | None): Name of the run file. If None, the dataset's dataset_id or file name will be used.
                Defaults to None.
            overwrite (bool): Whether to skip or overwrite already existing run files. Defaults to False.
        """
        super().__init__()
        self.save_dir = Path(save_dir) if save_dir is not None else None
        self.run_name = run_name
        self.overwrite = overwrite
        self.run_dfs: list[pd.DataFrame] = []

    def setup(self, trainer: Trainer, pl_module: LightningIRModule, stage: str) -> None:
        """Hook to setup the callback.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (LightningIRModule): LightningIR module.
            stage (str): Stage of the trainer, must be "test".
        Raises:
            ValueError: If the stage is not "test".
            ValueError: If no save_dir is provided and model_name_or_path is not a path (the model is not local).
        """
        if stage != "test":
            raise ValueError(f"{self.__class__.__name__} can only be used in test stage")
        if self.save_dir is None:
            default_save_dir = Path(pl_module.config.name_or_path)
            if default_save_dir.exists():
                self.save_dir = default_save_dir / "runs"
                print(f"Using default save_dir `{self.save_dir}` to save runs")
            else:
                raise ValueError("No save_dir provided and model_name_or_path is not a path")
        self._remove_overwrite_datasets(trainer, pl_module)

    def teardown(self, trainer: Trainer, pl_module: BiEncoderModule, stage: str) -> None:
        """Hook to cleanup the callback.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (LightningIRModule): LightningIR bi-encoder module used for indexing.
            stage (str): Stage of the trainer, must be "test".
        """
        self._cleanup(trainer, pl_module)

    def _get_save_path(self, pl_module: LightningIRModule, dataset: IRDataset) -> Path:
        if self.save_dir is None:
            raise ValueError("No save_dir found; call setup before using this method")
        if self.run_name is not None:
            run_file = self.run_name
        elif isinstance(dataset, QueryDataset):
            run_file = f"{dataset.dataset_id.replace('/', '-')}.run"
        elif isinstance(dataset, RunDataset):
            if dataset.run_path is None:
                run_file = f"{dataset.dataset_id.replace('/', '-')}.run"
            else:
                run_file = f"{dataset.run_path.name.split('.')[0]}.run"
        else:
            raise ValueError("Expected QueryDataset or RunDataset for ranking")
        run_file_path = self.save_dir / run_file
        return run_file_path

    def _rank(self, batch: RankBatch, output: LightningIROutput) -> tuple[torch.Tensor, list[str], list[int]]:
        scores = output.scores
        if scores is None:
            raise ValueError("Expected output to have scores")
        doc_ids = batch.doc_ids
        if doc_ids is None:
            raise ValueError("Expected batch to have doc_ids")
        scores = scores.view(-1)
        num_docs = [len(_doc_ids) for _doc_ids in doc_ids]
        doc_ids = list(itertools.chain.from_iterable(doc_ids))
        if scores.shape[0] != len(doc_ids):
            raise ValueError("scores and doc_ids must have the same length")
        return scores, doc_ids, num_docs

    def _write_run_dfs(self, trainer: Trainer, pl_module: LightningIRModule, dataloader_idx: int):
        if not trainer.is_global_zero or not self.run_dfs:
            return
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        dataset = pl_module.get_dataset(dataloader_idx)
        if dataset is None:
            raise ValueError("No dataset found to write run file")
        if not isinstance(dataset, (QueryDataset, RunDataset)):
            raise ValueError("Expected QueryDataset or RunDataset for ranking")
        run_file_path = self._get_save_path(pl_module, dataset)
        run_file_path.parent.mkdir(parents=True, exist_ok=True)
        run_df = pd.concat(self.run_dfs, ignore_index=True)
        run_df.to_csv(run_file_path, header=False, index=False, sep="\t", quoting=csv.QUOTE_NONE)

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningIRModule,
        outputs: LightningIROutput,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Hook to aggregate and write ranking to file.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (LightningIRModule): LightningIR module.
            outputs (LightningIROutput): Scored query documents pairs.
            batch (Any): Batch of input data.
            batch_idx (int): Index of batch in the current dataset.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        Raises:
            ValueError: If the batch does not have query_ids.
        """
        super().on_test_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
        batch = self._gather(pl_module, batch)
        outputs = self._gather(pl_module, outputs)
        if not trainer.is_global_zero:
            return

        query_ids = batch.query_ids
        if query_ids is None:
            raise ValueError("Expected batch to have query_ids")
        scores, doc_ids, num_docs = self._rank(batch, outputs)
        scores = scores.float().cpu().numpy()

        query_ids = list(
            itertools.chain.from_iterable(itertools.repeat(query_id, num) for query_id, num in zip(query_ids, num_docs))
        )
        run_df = pd.DataFrame(zip(query_ids, doc_ids, scores), columns=["query_id", "doc_id", "score"])
        run_df = run_df.sort_values(["query_id", "score"], ascending=[True, False])
        run_df["rank"] = run_df.groupby("query_id")["score"].rank(ascending=False, method="first").astype(int)
        run_df["q0"] = 0
        run_df["system"] = pl_module.model.__class__.__name__
        run_df = run_df[RUN_HEADER]

        self.run_dfs.append(run_df)

        if batch_idx == trainer.num_test_batches[dataloader_idx] - 1:
            self._write_run_dfs(trainer, pl_module, dataloader_idx)
            self.run_dfs = []


class SearchCallback(RankCallback, _IndexDirMixin):
    def __init__(
        self,
        search_config: SearchConfig,
        index_dir: Path | str | None = None,
        index_name: str | None = None,
        save_dir: Path | str | None = None,
        run_name: str | None = None,
        overwrite: bool = False,
        use_gpu: bool = True,
    ) -> None:
        """Callback to which uses index to retrieve documents efficiently.

        Args:
            search_config (SearchConfig): Configuration of the :py:class:`~lightning_ir.retrieve.base.searcher.Searcher`
            index_dir (Path | str | None): Directory where indexes are stored. Defaults to None.
            index_name (str | None): Name of the index. If None, the dataset's dataset_id or file name will be used.
                Defaults to None.
            save_dir (Path | str | None): Directory to save run files to. If None, run files will be saved in the
                model's directory. Defaults to None.
            run_name (str | None): Name of the run file. If None, the dataset's dataset_id or file name will be used.
                Defaults to None.
            overwrite (bool): Whether to skip or overwrite already existing run files. Defaults to False.
            use_gpu (bool): Toggle to use GPU for retrieval. Defaults to True.
        """
        super().__init__(save_dir=save_dir, run_name=run_name, overwrite=overwrite)
        self.search_config = search_config
        self.index_dir = index_dir
        self.index_name = index_name
        self.overwrite = overwrite
        self.use_gpu = use_gpu
        self.searcher: Searcher

    def _get_searcher(self, trainer: Trainer, pl_module: BiEncoderModule, dataset_idx: int) -> Searcher:
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        dataset = dataloaders[dataset_idx].dataset

        index_dir = self._get_index_dir(pl_module, dataset)
        if hasattr(self, "searcher"):
            if self.searcher.index_dir == index_dir:
                return self.searcher
            # free up memory
            del self.searcher
            gc.collect()
            torch.cuda.empty_cache()

        searcher = self.search_config.search_class(index_dir, self.search_config, pl_module, self.use_gpu)
        return searcher

    def _rank(
        self, batch: SearchBatch | RankBatch, output: LightningIROutput
    ) -> tuple[torch.Tensor, list[str], list[int]]:
        if batch.doc_ids is None:
            raise ValueError("BiEncoderModule did not return doc_ids when searching")
        dummy_docs = [[""] * len(ids) for ids in batch.doc_ids]
        batch = RankBatch(batch.queries, dummy_docs, batch.query_ids, batch.doc_ids, batch.qrels)
        return super()._rank(batch, output)

    def on_test_start(self, trainer: Trainer, pl_module: BiEncoderModule) -> None:
        """Hook to validate datasets

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module.
        Raises:
            ValueError: If no test_dataloaders are found.
            ValueError: If not all test datasets are :py:class:`~lightning_ir.data.dataset.QueryDataset`.
        """
        dataloaders = trainer.test_dataloaders
        if dataloaders is None:
            raise ValueError("No test_dataloaders found")
        datasets = [dataloader.dataset for dataloader in dataloaders]
        if not all(isinstance(dataset, (QueryDataset, _DummyIterableDataset)) for dataset in datasets):
            raise ValueError("Expected QueryDatasets for indexing")

    def on_test_batch_start(
        self, trainer: Trainer, pl_module: BiEncoderModule, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Hook to initialize searcher for new datasets.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (BiEncoderModule): LightningIR bi-encoder module.
            batch (Any): Batch of input data.
            batch_idx (int): Index of the batch in the dataset.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        """
        if batch_idx == 0:
            self.searcher = self._get_searcher(trainer, pl_module, dataloader_idx)
            pl_module.searcher = self.searcher
        super().on_test_batch_start(trainer, pl_module, batch, batch_idx, dataloader_idx)


class ReRankCallback(RankCallback):
    pass


class RegisterLocalDatasetCallback(Callback):
    def __init__(
        self,
        dataset_id: str,
        docs: str | None = None,
        queries: str | None = None,
        qrels: str | None = None,
        docpairs: str | None = None,
        scoreddocs: str | None = None,
        qrels_defs: dict[int, str] | None = None,
    ):
        """Registers a local dataset with ``ir_datasets``. After registering the dataset, it can be loaded using
        ``ir_datasets.load(dataset_id)``. Currently, the following (optionally gzipped) file types are supported:

        - ``.tsv``, ``.json``, or ``.jsonl`` for documents and queries
        - ``.tsv`` or ``.qrels`` for qrels
        - ``.tsv`` for training n-tuples
        - ``.tsv`` or ``.run`` for scored documents / run files

        Args:
            dataset_id (str): Dataset id.
            docs (str | None): Path to documents file or valid ir_datasets id from which documents should be taken.
                Defaults to None.
            queries (str | None): Path to queries file or valid ir_datasets id from which queries should be taken.
                Defaults to None.
            qrels (str | None): Path to qrels file or valid ir_datasets id from which qrels will be taken.
                Defaults to None.
            docpairs (str | None): Path to training n-tuple file or valid ir_datasets id from which training tuples
                will be taken. Defaults to None.
            scoreddocs (str | None): Path to run file or valid ir_datasets id from which scored documents will be taken.
                Defaults to None.
            qrels_defs (dict[int, str] | None): Optional dictionary describing the relevance levels of the qrels.
                Defaults to None.
        """
        super().__init__()
        self.dataset_id = dataset_id
        self.docs = docs
        self.queries = queries
        self.qrels = qrels
        self.docpairs = docpairs
        self.scoreddocs = scoreddocs
        self.qrels_defs = qrels_defs

    def setup(self, trainer: Trainer, pl_module: LightningIRModule, stage: str) -> None:
        """Hook that registers dataset.

        Args:
            trainer (Trainer): PyTorch Lightning Trainer.
            pl_module (LightningIRModule): Lightning IR module.
            stage (str): Stage of the trainer.
        """
        register_new_dataset(
            self.dataset_id,
            docs=self.docs,
            queries=self.queries,
            qrels=self.qrels,
            docpairs=self.docpairs,
            scoreddocs=self.scoreddocs,
            qrels_defs=self.qrels_defs,
        )



import pytest
from transformers import AutoTokenizer

from lightning_ir.base.tokenizer import LightningIRTokenizerClassFactory
from lightning_ir.bi_encoder.bi_encoder_config import BiEncoderConfig
from lightning_ir.cross_encoder.cross_encoder_config import CrossEncoderConfig


def test_serialize_deserialize(
    model_name_or_path: str,
    config: BiEncoderConfig | CrossEncoderConfig,
    tmpdir_factory: pytest.TempdirFactory,
):
    save_dir = tmpdir_factory.mktemp(config.model_type)
    Tokenizer = LightningIRTokenizerClassFactory(config.__class__).from_pretrained(model_name_or_path)
    tokenizer_kwargs = config.get_tokenizer_kwargs(Tokenizer)
    tokenizer = Tokenizer.from_pretrained(model_name_or_path, **tokenizer_kwargs)
    tokenizer.save_pretrained(save_dir)
    new_tokenizers = [
        AutoTokenizer.from_pretrained(save_dir),
        Tokenizer.from_pretrained(save_dir),
        Tokenizer.__bases__[0].from_pretrained(save_dir),
    ]
    for new_tokenizer in new_tokenizers:
        assert new_tokenizer.__class__.__name__ == tokenizer.__class__.__name__
        for key in tokenizer_kwargs:
            assert getattr(tokenizer, key) == getattr(new_tokenizer, key)


def test_bi_encoder_tokenizer(
    bi_encoder_config: BiEncoderConfig, model_name_or_path: str, monkeypatch: pytest.MonkeyPatch
):
    monkeypatch.setattr(bi_encoder_config, "add_marker_tokens", True)

    Tokenizer = LightningIRTokenizerClassFactory(type(bi_encoder_config)).from_pretrained(model_name_or_path)
    tokenizer = Tokenizer.from_pretrained(model_name_or_path, config=bi_encoder_config)

    query = "What is the capital of France?"
    doc = "Paris is the capital of France."
    encoding = tokenizer.tokenize(query, doc)
    assert encoding is not None
    query_encoding = encoding["query_encoding"]
    doc_encoding = encoding["doc_encoding"]
    assert query_encoding.input_ids[1] == tokenizer.query_token_id
    assert doc_encoding.input_ids[1] == tokenizer.doc_token_id

    query = ["What is the capital of France?"]
    doc = ["Paris is the capital of France."]
    encoding = tokenizer.tokenize(query, doc)
    assert encoding is not None
    query_encoding = encoding["query_encoding"]
    doc_encoding = encoding["doc_encoding"]
    assert query_encoding.input_ids[0][1] == tokenizer.query_token_id
    assert doc_encoding.input_ids[0][1] == tokenizer.doc_token_id


def test_cross_encoder_tokenizer(cross_encoder_config: CrossEncoderConfig, model_name_or_path: str):
    Tokenizer = LightningIRTokenizerClassFactory(type(cross_encoder_config)).from_pretrained(model_name_or_path)
    tokenizer = Tokenizer.from_pretrained(model_name_or_path, query_length=2, doc_length=4)

    query = "What is the capital of France?"
    doc = "Paris is the capital of France."
    encoding = tokenizer.tokenize(query, doc)["encoding"]
    assert encoding is not None
    assert len(encoding.input_ids[0]) == tokenizer.query_length + tokenizer.doc_length + 3

    query = ["What is the capital of France?"]
    doc = ["Paris is the capital of France."]
    encoding = tokenizer.tokenize(query, doc)["encoding"]
    assert encoding is not None
    assert len(encoding.input_ids[0]) == tokenizer.query_length + tokenizer.doc_length + 3

    query = "What is the capital of France?"
    doc = ["Paris is the capital of France."]
    encoding = tokenizer.tokenize(query, doc)["encoding"]
    assert encoding is not None
    assert len(encoding.input_ids[0]) == tokenizer.query_length + tokenizer.doc_length + 3

    query = ["What is the capital of France?"]
    doc = "Paris is the capital of France."
    with pytest.raises(ValueError):
        encoding = tokenizer.tokenize(query, doc)["encoding"]



"""
DataModule for Lightning IR that handles batching and collation of data samples.

This module defines the LightningIRDataModule class that handles batching and collation of data samples for training and
inference in Lightning IR.
"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import Any, Literal

import torch
from lightning import LightningDataModule
from torch.utils.data import DataLoader, IterableDataset

from .data import IndexBatch, RankBatch, SearchBatch, TrainBatch
from .dataset import (
    DocDataset,
    DocSample,
    QueryDataset,
    QuerySample,
    RankSample,
    RunDataset,
    TupleDataset,
    _DummyIterableDataset,
)


class LightningIRDataModule(LightningDataModule):
    def __init__(
        self,
        train_dataset: RunDataset | TupleDataset | None = None,
        train_batch_size: int | None = None,
        shuffle_train: bool = True,
        inference_datasets: Sequence[RunDataset | TupleDataset | QueryDataset | DocDataset] | None = None,
        inference_batch_size: int | None = None,
        num_workers: int = 0,
    ) -> None:
        """Initializes a new Lightning IR DataModule.

        Args:
            train_dataset (RunDataset | TupleDataset | None): A training dataset. Defaults to None.
            train_batch_size (int | None): Batch size to use for training. Defaults to None.
            shuffle_train (bool): Whether to shuffle the training data. Defaults to True.
            inference_datasets (Sequence[RunDataset | TupleDataset | QueryDataset | DocDataset] | None): list of
                datasets to use for inference (indexing, searching, and re-ranking). Defaults to None.
            inference_batch_size (int | None): Batch size to use for inference. Defaults to None.
            num_workers (int): Number of workers for loading data in parallel. Defaults to 0.
        """
        super().__init__()
        self.num_workers = num_workers

        self.train_dataset = train_dataset
        self.train_batch_size = train_batch_size
        self.shuffle_train = shuffle_train
        self.inference_datasets = None if inference_datasets is None else list(inference_datasets)
        self.inference_batch_size = inference_batch_size

        if (self.train_batch_size is not None) != (self.train_dataset is not None):
            raise ValueError("Both train_batch_size and train_dataset must be provided.")
        if (self.inference_batch_size is not None) != (self.inference_datasets is not None):
            raise ValueError("Both inference_batch_size and inference_dataset must be provided.")

    def _setup_inference(self, stage: Literal["validate", "test"]) -> None:
        if self.inference_datasets is None:
            return
        for inference_dataset in self.inference_datasets:
            if isinstance(inference_dataset, TupleDataset):
                if stage == "test":
                    raise ValueError("Prediction cannot be performed with TupleDataset.")
            elif isinstance(inference_dataset, RunDataset):
                if inference_dataset.sampling_strategy == "single_relevant":
                    raise ValueError("Inference RunDataset cannot use the single_relevant sampling strategy.")
            elif isinstance(inference_dataset, (QueryDataset, DocDataset)):
                pass
            else:
                raise ValueError(
                    "Inference Dataset must be of type RunDataset, TupleDataset, QueryDataset, or DocDataset."
                )

    def prepare_data(self) -> None:
        """Downloads the data using ir_datasets if needed."""
        if self.train_dataset is not None:
            self.train_dataset.prepare_data()
        if self.inference_datasets is not None:
            for inference_dataset in self.inference_datasets:
                inference_dataset.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test"]) -> None:
        """Sets up the data module for a given stage.

        Args:
            stage (Literal["fit", "validate", "test"]): Stage to set up the data module for.
        Raises:
            ValueError: If the stage is `fit` and no training dataset is provided.
        """
        if stage == "fit":
            if self.train_dataset is None:
                raise ValueError("A training dataset and config must be provided.")
        if stage == "fit":
            stage = "validate"
        self._setup_inference(stage)

    def train_dataloader(self) -> DataLoader:
        """Returns a dataloader for training.

        Returns:
            DataLoader: Dataloader for training.
        Raises:
            ValueError: If no training dataset is found.
        """
        if self.train_dataset is None:
            raise ValueError("No training dataset found.")
        return DataLoader(
            self.train_dataset,
            batch_size=self.train_batch_size,
            num_workers=self.num_workers,
            collate_fn=self._collate_fn,
            shuffle=(False if isinstance(self.train_dataset, IterableDataset) else self.shuffle_train),
            prefetch_factor=16 if self.num_workers > 0 else None,
        )

    def val_dataloader(self) -> list[DataLoader]:
        """Returns a list of dataloaders for validation.

        Returns:
            list[DataLoader]: Dataloaders for validation.
        """
        return self.inference_dataloader()

    def test_dataloader(self) -> list[DataLoader]:
        """Returns a list of dataloaders for testing.

        Returns:
            list[DataLoader]: Dataloaders for testing.
        """
        return self.inference_dataloader()

    def predict_dataloader(self) -> Any:
        """Returns a list of dataloaders for predicting.

        Returns:
            list[DataLoader]: Dataloaders for predicting.
        """
        return self.inference_dataloader()

    def inference_dataloader(self) -> list[DataLoader]:
        """Returns a list of dataloaders for inference (validation, testing, or predicting).

        Returns:
            list[DataLoader]: Dataloaders for inference.
        """
        inference_datasets = self.inference_datasets or []
        dataloaders = [
            DataLoader(
                dataset,
                batch_size=self.inference_batch_size,
                num_workers=self.num_workers,
                collate_fn=self._collate_fn,
                prefetch_factor=16 if self.num_workers > 0 else None,
            )
            for dataset in inference_datasets
            if not dataset._SKIP
        ]
        if not dataloaders:
            dataloaders = [DataLoader(_DummyIterableDataset())]
        return dataloaders

    def _aggregate_samples(self, samples: Sequence[RankSample | QuerySample | DocSample]) -> dict[str, Any]:
        aggregated = defaultdict(list)
        field_options = {
            "query_id": {"extend": False},
            "query": {"extend": False},
            "doc_id": {"extend": False},
            "doc_ids": {"extend": False},
            "doc": {"extend": False},
            "docs": {"extend": False},
            "targets": {"extend": True},
            "qrels": {"extend": True},
        }
        for sample in samples:
            for field in sample.__dict__:
                extend = field_options[field]["extend"]
                key = field if field.endswith("s") else f"{field}s"
                value = getattr(sample, field)
                if value is None:
                    continue
                if extend:
                    aggregated[key].extend(value)
                else:
                    aggregated[key].append(value)
        return aggregated

    def _clean_sample(self, aggregated: dict[str, Any]) -> dict[str, Any]:
        kwargs: dict[str, Any] = dict(aggregated)
        if "querys" in kwargs:
            kwargs["queries"] = kwargs["querys"]
            del kwargs["querys"]
        if "targets" in kwargs:
            kwargs["targets"] = torch.stack(kwargs["targets"])
        return kwargs

    def _parse_batch(
        self, sample: RankSample | QuerySample | DocSample, **kwargs
    ) -> RankBatch | TrainBatch | IndexBatch | SearchBatch:
        if isinstance(sample, RankSample):
            if "targets" in kwargs:
                return TrainBatch(**kwargs)
            else:
                return RankBatch(**kwargs)
        if isinstance(sample, QuerySample):
            return SearchBatch(**kwargs)
        if isinstance(sample, DocSample):
            return IndexBatch(**kwargs)
        raise ValueError("Invalid dataset configuration.")

    def _collate_fn(
        self,
        samples: Sequence[RankSample | QuerySample | DocSample] | RankSample | QuerySample | DocSample,
    ) -> TrainBatch | RankBatch | IndexBatch | SearchBatch:
        if isinstance(samples, (RankSample, QuerySample, DocSample)):
            samples = [samples]
        aggregated = self._aggregate_samples(samples)
        kwargs = self._clean_sample(aggregated)
        return self._parse_batch(samples[0], **kwargs)



"""
Model module for cross-encoder models.

This module defines the model class used to implement cross-encoder models.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass

import torch
from transformers import BatchEncoding

from ..base import LightningIRModel, LightningIROutput
from ..base.model import batch_encoding_wrapper
from . import CrossEncoderConfig


@dataclass
class CrossEncoderOutput(LightningIROutput):
    """Dataclass containing the output of a cross-encoder model"""

    embeddings: torch.Tensor | None = None
    """Joint query-document embeddings"""


class CrossEncoderModel(LightningIRModel, ABC):
    config_class: type[CrossEncoderConfig] = CrossEncoderConfig
    """Configuration class for cross-encoder models."""

    def __init__(self, config: CrossEncoderConfig, *args, **kwargs):
        """A cross-encoder model that jointly encodes a query and document(s). The contextualized embeddings are
        aggragated into a single vector and fed to a linear layer which computes a final relevance score.

        Args:
            config (CrossEncoderConfig): Configuration for the cross-encoder model.
        """
        super().__init__(config, *args, **kwargs)
        self.config: CrossEncoderConfig

    @batch_encoding_wrapper
    @abstractmethod
    def forward(self, encoding: BatchEncoding) -> CrossEncoderOutput:
        """Computes contextualized embeddings for the joint query-document input sequence and computes a relevance
        score.

        Args:
            encoding (BatchEncoding): Tokenizer encoding for the joint query-document input sequence.
        Returns:
            CrossEncoderOutput: Output of the model.
        """
        pass



"""
Base classes and abstract interfaces for loss functions in the Lightning IR framework.

This module defines the abstract base classes and common functionality for all loss functions
used in the Lightning IR framework.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

import torch

if TYPE_CHECKING:
    from ..base import LightningIROutput
    from ..bi_encoder import BiEncoderOutput
    from ..data import TrainBatch


class LossFunction(ABC):
    """Base class for loss functions in the Lightning IR framework."""

    @abstractmethod
    def compute_loss(self, output: LightningIROutput, *args, **kwargs) -> torch.Tensor:
        """Compute the loss for the given output.

        Args:
            output (LightningIROutput): The output from the model.
        Returns:
            torch.Tensor: The computed loss.
        """
        ...

    def process_scores(self, output: LightningIROutput) -> torch.Tensor:
        """Process the scores from the output.

        Args:
            output (LightningIROutput): The output from the model.
        Returns:
            torch.Tensor: The scores tensor.
        """
        if output.scores is None:
            raise ValueError("Expected scores in LightningIROutput")
        return output.scores

    def process_targets(self, scores: torch.Tensor, batch: TrainBatch) -> torch.Tensor:
        """Process the targets from the batch.

        Args:
            scores (torch.Tensor): The scores tensor.
            batch (TrainBatch): The training batch.
        Returns:
            torch.Tensor: The processed targets tensor.
        """
        targets = batch.targets
        if targets is None:
            raise ValueError("Expected targets in TrainBatch")
        if targets.ndim > scores.ndim:
            return targets.amax(-1)
        return targets


class ScoringLossFunction(LossFunction):
    """Base class for loss functions that operate on scores."""

    @abstractmethod
    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the loss based on the scores and targets in the output and batch.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        ...


class EmbeddingLossFunction(LossFunction):
    """Base class for loss functions that operate on embeddings."""

    @abstractmethod
    def compute_loss(self, output: BiEncoderOutput) -> torch.Tensor:
        """Compute the loss based on the embeddings in the output.

        Args:
            output (BiEncoderOutput): The output from the model containing query and document embeddings.
        Returns:
            torch.Tensor: The computed loss.
        """
        ...


class RegularizationLossFunction(EmbeddingLossFunction):
    """Base class for regularization loss functions that operate on embeddings."""

    def __init__(self, query_weight: float = 1e-4, doc_weight: float = 1e-4) -> None:
        """Initialize the RegularizationLossFunction.

        Args:
            query_weight (float): Weight for the query embeddings regularization. Defaults to 1e-4.
            doc_weight (float): Weight for the document embeddings regularization. Defaults to 1e-4.
        """
        self.query_weight = query_weight
        self.doc_weight = doc_weight

    def process_embeddings(self, output: BiEncoderOutput) -> tuple[torch.Tensor, torch.Tensor]:
        """Process the embeddings from the output.

        Args:
            output (BiEncoderOutput): The output from the model containing query and document embeddings.
        Returns:
            tuple[torch.Tensor, torch.Tensor]: The processed query and document embeddings.
        Raises:
            ValueError: If query_embeddings are not present in the output.
            ValueError: If doc_embeddings are not present in the output.
        """
        query_embeddings = output.query_embeddings
        doc_embeddings = output.doc_embeddings
        if query_embeddings is None:
            raise ValueError("Expected query_embeddings in BiEncoderOutput")
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")
        return query_embeddings.embeddings, doc_embeddings.embeddings


class PairwiseLossFunction(ScoringLossFunction):
    """Base class for pairwise loss functions."""

    def get_pairwise_idcs(self, targets: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Get pairwise indices for positive and negative samples based on targets.

        Args:
            targets (torch.Tensor): The targets tensor containing relevance labels.
        Returns:
            tuple[torch.Tensor, ...]: Indices of positive and negative samples.
        """
        # positive items are items where label is greater than other label in sample
        return torch.nonzero(targets[..., None] > targets[:, None], as_tuple=True)


class ListwiseLossFunction(ScoringLossFunction):
    """Base class for listwise loss functions."""

    pass


class InBatchLossFunction(LossFunction):
    """Base class for in-batch loss functions that compute in-batch indices for positive and negative samples."""

    def __init__(
        self,
        pos_sampling_technique: Literal["all", "first"] = "all",
        neg_sampling_technique: Literal["all", "first", "all_and_non_first"] = "all",
        max_num_neg_samples: int | None = None,
    ):
        """Initialize the InBatchLossFunction.

        Args:
            pos_sampling_technique (Literal["all", "first"]): Technique for positive sample sampling.
            neg_sampling_technique (Literal["all", "first", "all_and_non_first"]): Technique for negative sample
                sampling.
            max_num_neg_samples (int | None): Maximum number of negative samples to consider. If None, all negative
                samples are considered.
        Raises:
            ValueError: If the negative sampling technique is invalid for the given positive sampling technique.
        """
        super().__init__()
        self.pos_sampling_technique = pos_sampling_technique
        self.neg_sampling_technique = neg_sampling_technique
        self.max_num_neg_samples = max_num_neg_samples
        if self.neg_sampling_technique == "all_and_non_first" and self.pos_sampling_technique != "first":
            raise ValueError("all_and_non_first is only valid with pos_sampling_technique first")

    def _get_pos_mask(
        self,
        num_queries: int,
        num_docs: int,
        max_idx: torch.Tensor,
        min_idx: torch.Tensor,
        output: LightningIROutput,
        batch: TrainBatch,
    ) -> torch.Tensor:
        """Get the mask for positive samples based on the sampling technique.

        Args:
            num_queries (int): Number of queries in the batch.
            num_docs (int): Number of documents per query.
            max_idx (torch.Tensor): Maximum index for each query.
            min_idx (torch.Tensor): Minimum index for each query.
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: A mask tensor indicating the positions of positive samples.
        Raises:
            ValueError: If the positive sampling technique is invalid.
        """
        if self.pos_sampling_technique == "all":
            pos_mask = torch.arange(num_queries * num_docs)[None].greater_equal(min_idx) & torch.arange(
                num_queries * num_docs
            )[None].less(max_idx)
        elif self.pos_sampling_technique == "first":
            pos_mask = torch.arange(num_queries * num_docs)[None].eq(min_idx)
        else:
            raise ValueError("invalid pos sampling technique")
        return pos_mask

    def _get_neg_mask(
        self,
        num_queries: int,
        num_docs: int,
        max_idx: torch.Tensor,
        min_idx: torch.Tensor,
        output: LightningIROutput,
        batch: TrainBatch,
    ) -> torch.Tensor:
        """Get the mask for negative samples based on the sampling technique.

        Args:
            num_queries (int): Number of queries in the batch.
            num_docs (int): Number of documents per query.
            max_idx (torch.Tensor): Maximum index for each query.
            min_idx (torch.Tensor): Minimum index for each query.
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: A mask tensor indicating the positions of negative samples.
        Raises:
            ValueError: If the negative sampling technique is invalid.
        """
        if self.neg_sampling_technique == "all_and_non_first":
            neg_mask = torch.arange(num_queries * num_docs)[None].not_equal(min_idx)
        elif self.neg_sampling_technique == "all":
            neg_mask = torch.arange(num_queries * num_docs)[None].less(min_idx) | torch.arange(num_queries * num_docs)[
                None
            ].greater_equal(max_idx)
        elif self.neg_sampling_technique == "first":
            neg_mask = torch.arange(num_queries * num_docs)[None, None].eq(min_idx).any(1) & torch.arange(
                num_queries * num_docs
            )[None].ne(min_idx)
        else:
            raise ValueError("invalid neg sampling technique")
        return neg_mask

    def get_ib_idcs(self, output: LightningIROutput, batch: TrainBatch) -> tuple[torch.Tensor, torch.Tensor]:
        """Get in-batch indices for positive and negative samples.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            tuple[torch.Tensor, torch.Tensor]: Indices of positive and negative samples.
        Raises:
            ValueError: If scores are not present in the output.
        """
        if output.scores is None:
            raise ValueError("Expected scores in LightningIROutput")
        num_queries, num_docs = output.scores.shape
        min_idx = torch.arange(num_queries)[:, None] * num_docs
        max_idx = min_idx + num_docs
        pos_mask = self._get_pos_mask(num_queries, num_docs, max_idx, min_idx, output, batch)
        neg_mask = self._get_neg_mask(num_queries, num_docs, max_idx, min_idx, output, batch)
        pos_idcs = pos_mask.nonzero(as_tuple=True)[1]
        neg_idcs = neg_mask.nonzero(as_tuple=True)[1]
        if self.max_num_neg_samples is not None:
            neg_idcs = neg_idcs.view(num_queries, -1)
            if neg_idcs.shape[-1] > 1:
                neg_idcs = neg_idcs[:, torch.randperm(neg_idcs.shape[-1])]
            neg_idcs = neg_idcs[:, : self.max_num_neg_samples]
            neg_idcs = neg_idcs.reshape(-1)
        return pos_idcs, neg_idcs



from .colbert import register_colbert_docpairs
from .cross_architecture_knowledge_distillation import register_kd_docpairs
from .ir_datasets_utils import register_new_dataset
from .rank_distillm import register_rank_distillm
from .sbert import register_sbert_docpairs


def _register_external_datasets():
    register_colbert_docpairs()
    register_kd_docpairs()
    register_rank_distillm()
    register_sbert_docpairs()


__all__ = ["register_new_dataset"]



import torch
from transformers.activations import get_activation
from transformers.models.bert.modeling_bert import BertOnlyMLMHead
from transformers.models.distilbert.configuration_distilbert import DistilBertConfig


class DistilBertOnlyMLMHead(torch.nn.Module):
    def __init__(self, config: DistilBertConfig) -> None:
        super().__init__()
        self.activation = get_activation(config.activation)
        self.vocab_transform = torch.nn.Linear(config.dim, config.dim)
        self.vocab_layer_norm = torch.nn.LayerNorm(config.dim, eps=1e-12)
        self.vocab_projector = torch.nn.Linear(config.dim, config.vocab_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.vocab_transform(x)
        x = self.activation(x)
        x = self.vocab_layer_norm(x)
        x = self.vocab_projector(x)
        return x


MODEL_TYPE_TO_LM_HEAD = {
    "bert": BertOnlyMLMHead,
    "distilbert": DistilBertOnlyMLMHead,
}

MODEL_TYPE_TO_STATE_DICT_KEY_MAPPING = {
    "bert": {"cls": "bert.projection"},
    "distilbert": {
        "vocab_transform": "distilbert.projection.vocab_transform",
        "vocab_layer_norm": "distilbert.projection.vocab_layer_norm",
        "vocab_projector": "distilbert.projection.vocab_projector",
    },
}

# NOTE: In the output embeddings and tied weight keys the cls key has already been unified and replaced by the
# projection key

MODEL_TYPE_TO_OUTPUT_EMBEDDINGS = {
    "bert": "predictions.decoder",
    "distilbert": "vocab_projector",
}

MODEL_TYPE_TO_TIED_WEIGHTS_KEYS = {
    "bert": ["predictions.decoder.bias", "predictions.decoder.weight"],
    "distilbert": ["vocab_projector.bias", "vocab_projector.weight"],
}



"""Plaid Indexer using fast-plaid library for Lightning IR Framework"""

from pathlib import Path

import torch

from ...bi_encoder import BiEncoderModule, BiEncoderOutput
from ...data import IndexBatch
from ...models import ColConfig
from ..base import IndexConfig, Indexer


class PlaidIndexer(Indexer):
    """Indexer for Plaid using fast-plaid library."""

    def __init__(
        self,
        index_dir: Path,
        index_config: "PlaidIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the PlaidIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (PlaidIndexConfig): Configuration for the Plaid indexer.
            module (BiEncoderModule): The BiEncoder module used for indexing.
            verbose (bool): Whether to print verbose output during indexing. Defaults to False.
        """
        super().__init__(index_dir, index_config, module, verbose)
        self.index_config: PlaidIndexConfig
        self.index = None
        self.embeddings = []

    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None:
        """Add embeddings from the index batch to the Plaid index.

        Args:
            index_batch (IndexBatch): Batch of data containing embeddings to be indexed.
            output (BiEncoderOutput): Output from the BiEncoder module containing embeddings.
        Raises:
            ValueError: If the output does not contain document embeddings.
        """
        doc_embeddings = output.doc_embeddings
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")

        if doc_embeddings.scoring_mask is None:
            doc_lengths = torch.ones(
                doc_embeddings.embeddings.shape[0], device=doc_embeddings.device, dtype=torch.int32
            )
            embeddings = doc_embeddings.embeddings[:, 0]
        else:
            doc_lengths = doc_embeddings.scoring_mask.sum(dim=1)
            embeddings = doc_embeddings.embeddings[doc_embeddings.scoring_mask]
        doc_ids = index_batch.doc_ids

        self.num_embeddings += embeddings.shape[0]
        self.num_docs += len(doc_ids)

        self.doc_lengths.extend(doc_lengths.int().cpu().tolist())
        self.doc_ids.extend(doc_ids)

        self.embeddings.extend(embeddings.cpu().split(doc_lengths.int().cpu().tolist()))

    def save(self) -> None:
        """Save the index configuration and document IDs to the index directory."""
        from fast_plaid import search

        index = search.FastPlaid(index=str(self.index_dir))

        index.create(
            documents_embeddings=self.embeddings,
            kmeans_niters=self.index_config.k_means_iters,
            nbits=self.index_config.n_bits,
            seed=self.index_config.seed,
        )

        super().save()


class PlaidIndexConfig(IndexConfig):
    """Configuration class for Plaid indexers in the Lightning IR framework."""

    indexer_class = PlaidIndexer
    SUPPORTED_MODELS = {ColConfig.model_type}

    def __init__(
        self,
        num_centroids: int,
        k_means_iters: int = 4,
        n_bits: int = 2,
        seed: int = 42,
    ) -> None:
        """Initialize the PlaidIndexConfig.

        Args:
            num_centroids (int): Number of centroids for the Plaid index.
            num_train_embeddings (int | None): Number of embeddings to use for training the index. If None, it will
                be set later. Defaults to None.
            k_means_iters (int): Number of iterations for k-means clustering. Defaults to 4.
            n_bits (int): Number of bits for the residual codec. Defaults to 2.
            seed (int): Random seed for reproducibility. Defaults to 42.
        """
        super().__init__()
        self.num_centroids = num_centroids
        self.k_means_iters = k_means_iters
        self.n_bits = n_bits
        self.seed = seed



from pathlib import Path

from transformers import AutoConfig

from lightning_ir.base.config import LightningIRConfig
from lightning_ir.base.module import LightningIRModule


def test_serialize_deserialize(module: LightningIRModule, tmp_path: Path):
    config = module.model.config
    config_class = module.model.config_class
    save_dir = str(tmp_path / config_class.model_type)
    config.save_pretrained(save_dir)
    new_configs = [
        config.__class__.from_pretrained(save_dir),
        config.__class__.__bases__[0].from_pretrained(save_dir),
        LightningIRConfig.from_pretrained(save_dir),
        AutoConfig.from_pretrained(save_dir),
    ]
    for new_config in new_configs:
        for key, value in config.__dict__.items():
            if key in (
                "torch_dtype",
                "_name_or_path",
                "_commit_hash",
                "transformers_version",
                "_attn_implementation_autoset",
                "_attn_implementation_internal",
            ):
                continue
            assert getattr(new_config, key) == value



import warnings

import torch
from huggingface_hub import hf_hub_download
from safetensors.torch import load_file
from transformers import T5EncoderModel

from ..base import (
    BACKBONE_MAPPING,
    CHECKPOINT_MAPPING,
    POST_LOAD_CALLBACKS,
    STATE_DICT_KEY_MAPPING,
    LightningIRModel,
)
from ..models import (
    CoilConfig,
    ColConfig,
    DprConfig,
    MonoConfig,
    SpladeConfig,
    UniCoilConfig,
)


def _map_colbert_marker_tokens(model: LightningIRModel) -> LightningIRModel:
    config = model.config
    query_token_id = config.vocab_size
    doc_token_id = config.vocab_size + 1
    model.resize_token_embeddings(config.vocab_size + 2, 8)
    embeddings = model.embeddings.word_embeddings.weight.data
    embeddings[query_token_id] = embeddings[1]  # [unused0]
    embeddings[doc_token_id] = embeddings[2]  # [unused1]
    return model


def _map_moderncolbert_marker_tokens(model: LightningIRModel) -> LightningIRModel:
    config = model.config
    query_token_id = config.vocab_size
    doc_token_id = config.vocab_size + 1
    model.resize_token_embeddings(config.vocab_size + 2, 8)
    embeddings = model.embeddings.tok_embeddings.weight.data
    embeddings[query_token_id] = embeddings[50368]  # [unused0]
    embeddings[doc_token_id] = embeddings[50369]  # [unused1]

    path = hf_hub_download(model.config.name_or_path, filename="model.safetensors", subfolder="1_Dense")
    state_dict = load_file(path)
    state_dict["weight"] = state_dict.pop("linear.weight")
    model.projection.load_state_dict(state_dict)
    return model


def _map_xtr_weights(model: LightningIRModel) -> LightningIRModel:
    warnings.warn(
        "The above warning, that the linear layer is not initialized, is expected and can be ignored."
        "The weights are initialized separately.",
        stacklevel=2,
    )
    path = hf_hub_download(model.config.name_or_path, filename="pytorch_model.bin", subfolder="2_Dense")
    state_dict = torch.load(path, weights_only=True, map_location="cpu")
    state_dict["projection.weight"] = state_dict.pop("linear.weight")
    model.load_state_dict(state_dict, strict=False)
    return model


def _map_mono_t5_weights(model: LightningIRModel) -> LightningIRModel:
    # [1176, 6136] true, false
    warnings.warn(
        "The above warning, that the linear layer is not initialized, is expected and can be ignored."
        "The weights are initialized separately.",
        stacklevel=2,
    )
    model.linear.weight.data = model.shared.weight.data[[6136, 1176]]
    return model


def _map_rank_t5_weights(model: LightningIRModel) -> LightningIRModel:
    # 32089 <extra_id_10>
    warnings.warn(
        "The above warning, that the linear layer is not initialized, is expected and can be ignored."
        "The weights are initialized separately.",
        stacklevel=2,
    )
    model.linear.weight.data = model.shared.weight.data[[32089]]
    return model


def _map_coil_weights(model: LightningIRModel) -> LightningIRModel:
    path = hf_hub_download(model.config.name_or_path, filename="model.pt")
    state_dict = torch.load(path, map_location="cpu")
    state_dict["token_projection.weight"] = state_dict.pop("tok_proj.weight")
    state_dict["token_projection.bias"] = state_dict.pop("tok_proj.bias")
    state_dict["cls_projection.weight"] = state_dict.pop("cls_proj.weight")
    state_dict["cls_projection.bias"] = state_dict.pop("cls_proj.bias")
    model.load_state_dict(state_dict, strict=False)
    return model


def _map_opensearch_splade_weights(model: LightningIRModel) -> LightningIRModel:
    path = hf_hub_download(
        model.config.name_or_path, filename="model.safetensors", subfolder="query_0_SparseStaticEmbedding"
    )
    state_dict = load_file(path)
    state_dict["weight"] = state_dict.pop("weight").unsqueeze(-1)
    model.query_weights.load_state_dict(state_dict)
    return model


def _register_external_models():
    CHECKPOINT_MAPPING.update(
        {
            "colbert-ir/colbertv2.0": ColConfig(
                query_length=32,
                doc_length=184,
                add_marker_tokens=True,
                normalization_strategy="l2",
                query_expansion=True,
                doc_mask_scoring_tokens="punctuation",
            ),
            "lightonai/GTE-ModernColBERT-v1": ColConfig(
                query_length=32,
                doc_length=296,
                add_marker_tokens=True,
                normalization_strategy="l2",
                query_expansion=False,
                projection="linear_no_bias",
                doc_mask_scoring_tokens="punctuation",
            ),
            "naver/splade-v3": SpladeConfig(),
            "naver/splade-v3-distilbert": SpladeConfig(),
            "naver/splade-v3-doc": SpladeConfig(query_expansion=False, query_weighting=None),
            "naver/splade-v3-lexical": SpladeConfig(query_expansion=False),
            "naver/splade_v2_distil": SpladeConfig(),
            "opensearch-project/opensearch-neural-sparse-encoding-v2-distill": SpladeConfig(),
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v2-distill": SpladeConfig(
                query_expansion=False, query_weighting="static"
            ),
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v2-mini": SpladeConfig(
                query_expansion=False,
                query_weighting="static",
            ),
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v3-distill": SpladeConfig(
                query_expansion=False, query_weighting="static", sparsification_strategy="relu_2xlog"
            ),
            "sentence-transformers/msmarco-bert-base-dot-v5": DprConfig(projection=None, pooling_strategy="mean"),
            "sentence-transformers/msmarco-distilbert-dot-v5": DprConfig(projection=None, pooling_strategy="mean"),
            "sentence-transformers/msmarco-MiniLM-L-6-v3": DprConfig(projection=None, pooling_strategy="mean"),
            "castorini/monot5-base-msmarco-10k": MonoConfig(
                scoring_strategy="mono", query_length=None, doc_length=None
            ),
            "castorini/monot5-base-msmarco": MonoConfig(scoring_strategy="mono", query_length=None, doc_length=None),
            "castorini/monot5-large-msmarco-10k": MonoConfig(
                scoring_strategy="mono", query_length=None, doc_length=None
            ),
            "castorini/monot5-large-msmarco": MonoConfig(scoring_strategy="mono", query_length=None, doc_length=None),
            "castorini/monot5-3b-msmarco-10k": MonoConfig(scoring_strategy="mono", query_length=None, doc_length=None),
            "castorini/monot5-3b-msmarco": MonoConfig(scoring_strategy="mono", query_length=None, doc_length=None),
            "Soyoung97/RankT5-base": MonoConfig(scoring_strategy="rank", query_length=None, doc_length=None),
            "Soyoung97/RankT5-large": MonoConfig(scoring_strategy="rank", query_length=None, doc_length=None),
            "Soyoung97/RankT5-3b": MonoConfig(scoring_strategy="rank", query_length=None, doc_length=None),
            "castorini/monobert-large-msmarco-finetune-only": MonoConfig(
                scoring_strategy="mono", linear_bias=True, pooling_strategy="bert_pool"
            ),
            "castorini/monobert-large-msmarco": MonoConfig(
                scoring_strategy="mono", linear_bias=True, pooling_strategy="bert_pool"
            ),
            "fschlatt/coil-with-hn": CoilConfig(),
            "castorini/unicoil-noexp-msmarco-passage": UniCoilConfig(projection="linear"),
            "google/xtr-base-en": ColConfig(projection="linear_no_bias", normalization_strategy="l2"),
        }
    )
    BACKBONE_MAPPING.update({"google/xtr-base-en": T5EncoderModel})
    STATE_DICT_KEY_MAPPING.update(
        {
            "colbert-ir/colbertv2.0": [("linear.weight", "bert.projection.weight")],
            "castorini/monobert-large-msmarco-finetune-only": [
                ("classifier.weight", "bert.linear.weight"),
                ("classifier.bias", "bert.linear.bias"),
                ("bert.pooler.dense.weight", "bert.bert_pool.0.weight"),
                ("bert.pooler.dense.bias", "bert.bert_pool.0.bias"),
            ],
            "castorini/monobert-large-msmarco": [
                ("classifier.weight", "bert.linear.weight"),
                ("classifier.bias", "bert.linear.bias"),
                ("bert.pooler.dense.weight", "bert.bert_pool.0.weight"),
                ("bert.pooler.dense.bias", "bert.bert_pool.0.bias"),
            ],
            "castorini/unicoil-noexp-msmarco-passage": [
                ("coil_encoder.bert.", ""),
                ("coil_encoder.tok_proj", "token_projection"),
            ],
        }
    )
    POST_LOAD_CALLBACKS.update(
        {
            "colbert-ir/colbertv2.0": _map_colbert_marker_tokens,
            "lightonai/GTE-ModernColBERT-v1": _map_moderncolbert_marker_tokens,
            "castorini/monot5-base-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-base-msmarco": _map_mono_t5_weights,
            "castorini/monot5-large-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-large-msmarco": _map_mono_t5_weights,
            "castorini/monot5-3b-msmarco-10k": _map_mono_t5_weights,
            "castorini/monot5-3b-msmarco": _map_mono_t5_weights,
            "fschlatt/coil-with-hn": _map_coil_weights,
            "google/xtr-base-en": _map_xtr_weights,
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v2-distill": _map_opensearch_splade_weights,
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v2-mini": _map_opensearch_splade_weights,
            "opensearch-project/opensearch-neural-sparse-encoding-doc-v3-distill": _map_opensearch_splade_weights,
            "Soyoung97/RankT5-base": _map_rank_t5_weights,
            "Soyoung97/RankT5-large": _map_rank_t5_weights,
            "Soyoung97/RankT5-3b": _map_rank_t5_weights,
        }
    )



"""FAISS-based searcher for approximate nearest neighbor retrieval in the Lightning IR framework."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import torch

from ...bi_encoder.bi_encoder_model import BiEncoderEmbedding
from ...models import ColConfig, DprConfig
from ..base.packed_tensor import PackedTensor
from ..base.searcher import ApproximateSearchConfig, ApproximateSearcher

if TYPE_CHECKING:
    from ...bi_encoder import BiEncoderModule


class FaissSearcher(ApproximateSearcher):
    """FAISS-based searcher for approximate nearest neighbor retrieval in the Lightning IR framework."""

    def __init__(
        self,
        index_dir: Path | str,
        search_config: FaissSearchConfig,
        module: BiEncoderModule,
        use_gpu: bool = False,
    ) -> None:
        """Initialize the FaissSearcher.

        Args:
            index_dir (Path | str): Directory containing the FAISS index files.
            search_config (FaissSearchConfig): Configuration for the FAISS searcher.
            module (BiEncoderModule): The bi-encoder module used for embeddings.
            use_gpu (bool): Whether to use GPU for FAISS operations. Defaults to False.
        """
        import faiss

        self.search_config: FaissSearchConfig
        self.index = faiss.read_index(str(Path(index_dir) / "index.faiss"))
        if use_gpu and hasattr(faiss, "index_cpu_to_all_gpus"):
            self.index = faiss.index_cpu_to_all_gpus(self.index)
        ivf_index = None
        try:
            ivf_index = faiss.extract_index_ivf(self.index)
        except RuntimeError:
            pass
        if ivf_index is not None:
            ivf_index.nprobe = search_config.n_probe
            quantizer = getattr(ivf_index, "quantizer", None)
            if quantizer is not None:
                downcasted_quantizer = faiss.downcast_index(quantizer)
                hnsw = getattr(downcasted_quantizer, "hnsw", None)
                if hnsw is not None:
                    hnsw.efSearch = search_config.ef_search
        super().__init__(index_dir, search_config, module, use_gpu)

    def _candidate_retrieval(self, query_embeddings: BiEncoderEmbedding) -> tuple[PackedTensor, PackedTensor]:
        if query_embeddings.scoring_mask is None:
            embeddings = query_embeddings.embeddings[:, 0]
        else:
            embeddings = query_embeddings.embeddings[query_embeddings.scoring_mask]
        candidate_scores, candidate_idcs = self.index.search(embeddings.float().cpu(), self.search_config.candidate_k)
        candidate_scores = torch.from_numpy(candidate_scores).view(-1)
        candidate_idcs = torch.from_numpy(candidate_idcs).view(-1)
        num_candidates_per_query_vector = [self.search_config.candidate_k] * embeddings.shape[0]
        packed_candidate_scores = PackedTensor(candidate_scores, lengths=num_candidates_per_query_vector)
        packed_candidate_idcs = PackedTensor(candidate_idcs, lengths=num_candidates_per_query_vector)
        return packed_candidate_scores, packed_candidate_idcs

    def _gather_doc_embeddings(self, idcs: torch.Tensor) -> torch.Tensor:
        return torch.from_numpy(self.index.reconstruct_batch(idcs))


class FaissSearchConfig(ApproximateSearchConfig):
    """Configuration class for FAISS-based searchers in the Lightning IR framework."""

    search_class = FaissSearcher
    SUPPORTED_MODELS = {ColConfig.model_type, DprConfig.model_type}

    def __init__(
        self,
        k: int = 10,
        candidate_k: int = 100,
        imputation_strategy: Literal["min", "gather", "zero"] = "gather",
        n_probe: int = 1,
        ef_search: int = 16,
    ) -> None:
        """Initialize the FaissSearchConfig.

        Args:
            k (int): Number of top results to return. Defaults to 10.
            candidate_k (int): Number of candidates to retrieve before ranking. Defaults to 100.
            imputation_strategy (Literal["min", "gather", "zero"]): Strategy for handling missing scores.
                Defaults to "gather".
            n_probe (int): Number of probes for the IVF index. Defaults to 1.
            ef_search (int): Size of the dynamic list used during search. Defaults to 16.
        """
        super().__init__(k, candidate_k, imputation_strategy)
        self.n_probe = n_probe
        self.ef_search = ef_search



from __future__ import annotations

from typing import TYPE_CHECKING

import torch

from .base import EmbeddingLossFunction

if TYPE_CHECKING:
    from ..bi_encoder import BiEncoderOutput


class ContrastiveLocalLoss(EmbeddingLossFunction):
    """Loss function that computes a contrastive loss between a query and multiple document embeddings, such that only
    one document embedding has a a high similarity to the query embedding, while all other document embeddings
    have a low similarity. Originally proposed in:
    `Multi-View Document Representation Learning for Open-Domain Dense Retrieval \
    <https://aclanthology.org/2022.acl-long.414/>`_"""

    def __init__(self, temperature: float = 1.0) -> None:
        super().__init__()
        self.temperature = temperature

    def compute_loss(self, output: BiEncoderOutput) -> torch.Tensor:
        """Compute the loss based on the embeddings in the output.

        Args:
            output (BiEncoderOutput): The output from the model containing query and document embeddings.
        Returns:
            torch.Tensor: The computed loss.
        """
        similarity = output.similarity
        if similarity is None:
            raise ValueError("Expected similarity in BiEncoderOutput")
        targets = similarity.argmax(-1)
        loss = torch.nn.functional.cross_entropy(similarity, targets)
        return loss



from transformers import AutoConfig, AutoModel, AutoTokenizer

from ..bi_encoder import BiEncoderConfig, BiEncoderTokenizer
from ..cross_encoder import CrossEncoderConfig, CrossEncoderTokenizer
from . import (
    CoilConfig,
    CoilModel,
    ColConfig,
    ColModel,
    ColTokenizer,
    DprConfig,
    DprModel,
    MonoConfig,
    MonoModel,
    MvrConfig,
    MvrModel,
    MvrTokenizer,
    SetEncoderConfig,
    SetEncoderModel,
    SetEncoderTokenizer,
    SpladeConfig,
    SpladeModel,
    SpladeTokenizer,
    UniCoilConfig,
    UniCoilModel,
)


def _register_internal_models():
    AutoTokenizer.register(BiEncoderConfig, BiEncoderTokenizer)
    AutoTokenizer.register(CrossEncoderConfig, CrossEncoderTokenizer)
    AutoConfig.register(CoilConfig.model_type, CoilConfig)
    AutoModel.register(CoilConfig, CoilModel)
    AutoTokenizer.register(CoilConfig, BiEncoderTokenizer)
    AutoConfig.register(ColConfig.model_type, ColConfig)
    AutoModel.register(ColConfig, ColModel)
    AutoTokenizer.register(ColConfig, ColTokenizer)
    AutoConfig.register(DprConfig.model_type, DprConfig)
    AutoModel.register(DprConfig, DprModel)
    AutoTokenizer.register(DprConfig, BiEncoderTokenizer)
    AutoConfig.register(MonoConfig.model_type, MonoConfig)
    AutoModel.register(MonoConfig, MonoModel)
    AutoTokenizer.register(MonoConfig, CrossEncoderTokenizer)
    AutoConfig.register(MvrConfig.model_type, MvrConfig)
    AutoModel.register(MvrConfig, MvrModel)
    AutoTokenizer.register(MvrConfig, MvrTokenizer)
    AutoConfig.register(SetEncoderConfig.model_type, SetEncoderConfig)
    AutoModel.register(SetEncoderConfig, SetEncoderModel)
    AutoTokenizer.register(SetEncoderConfig, SetEncoderTokenizer)
    AutoConfig.register(SpladeConfig.model_type, SpladeConfig)
    AutoModel.register(SpladeConfig, SpladeModel)
    AutoTokenizer.register(SpladeConfig, SpladeTokenizer)
    AutoConfig.register(UniCoilConfig.model_type, UniCoilConfig)
    AutoModel.register(UniCoilConfig, UniCoilModel)
    AutoTokenizer.register(UniCoilConfig, BiEncoderTokenizer)



"""LightningModule for Lightning IR.

This module contains the main module class deriving from a LightningModule_.

.. _LightningModule: https://lightning.ai/docs/pytorch/stable/common/lightning_module.html
"""

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import pandas as pd
import torch
from lightning import LightningModule
from lightning.pytorch.trainer.states import RunningStage
from transformers import BatchEncoding, PreTrainedModel

from ..data import IRDataset, RankBatch, RunDataset, SearchBatch, TrainBatch
from ..loss import InBatchLossFunction, LossFunction
from .config import LightningIRConfig
from .model import LightningIRModel, LightningIROutput
from .tokenizer import LightningIRTokenizer
from .validation_utils import create_qrels_from_dicts, create_run_from_scores, evaluate_run


class LightningIRModule(LightningModule):
    """LightningIRModule base class. It dervies from a LightningModule_. LightningIRModules contain a
    LightningIRModel and a LightningIRTokenizer and implements the training, validation, and testing steps for the
    model. Derived classes must implement the forward method for the model.

    .. _LightningModule: https://lightning.ai/docs/pytorch/stable/common/lightning_module.html
    """

    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: LightningIRConfig | None = None,
        model: LightningIRModel | None = None,
        BackboneModel: type[PreTrainedModel] | None = None,
        loss_functions: Sequence[LossFunction | tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
        model_kwargs: Mapping[str, Any] | None = None,
    ):
        """Initializes the LightningIRModule.

        .. _ir-measures: https://ir-measur.es/en/latest/index.html

        Args:
            model_name_or_path (str | None): Name or path of backbone model or fine-tuned Lightning IR model.
                Defaults to None.
            config (LightningIRConfig | None): LightningIRConfig to apply when loading from backbone model.
                Defaults to None.
            model (LightningIRModel | None): Already instantiated Lightning IR model. Defaults to None.
            BackboneModel (type[PreTrainedModel] | None): Huggingface PreTrainedModel class to use as backbone
                instead of the default AutoModel. Defaults to None.
            loss_functions (Sequence[LossFunction | tuple[LossFunction, float]] | None):
                Loss functions to apply during fine-tuning, optional loss weights can be provided per loss function
                Defaults to None.
            evaluation_metrics (Sequence[str] | None): Metrics corresponding to ir-measures_ measure strings
                to apply during validation or testing. Defaults to None.
            model_kwargs (Mapping[str, Any] | None): Additional keyword arguments to pass to `from_pretrained` when
                loading a model. Defaults to None.
        Raises:
            ValueError: If both model and model_name_or_path are provided.
            ValueError: If neither model nor model_name_or_path are provided.
        """
        super().__init__()
        model_kwargs = model_kwargs if model_kwargs is not None else {}
        self.save_hyperparameters()
        if model is not None and model_name_or_path is not None:
            raise ValueError("Only one of model or model_name_or_path must be provided.")
        if model is None:
            if model_name_or_path is None:
                raise ValueError("Either model or model_name_or_path must be provided.")
            model = LightningIRModel.from_pretrained(
                model_name_or_path, config=config, BackboneModel=BackboneModel, **model_kwargs
            )

        self.model: LightningIRModel = model
        self.config = self.model.config
        self.loss_functions: list[tuple[LossFunction, float]] | None = None
        if loss_functions is not None:
            self.loss_functions = []
            for loss_function in loss_functions:
                if isinstance(loss_function, LossFunction):
                    self.loss_functions.append((loss_function, 1.0))
                else:
                    self.loss_functions.append(loss_function)
        self.evaluation_metrics = evaluation_metrics
        self._optimizer: torch.optim.Optimizer | None = None
        self.tokenizer = LightningIRTokenizer.from_pretrained(self.config.name_or_path, config=self.config)
        self._additional_log_metrics: dict[str, float] = {}

    def on_train_start(self) -> None:
        """Called at the beginning of training after sanity check."""
        super().on_train_start()
        # NOTE huggingface models are in eval mode by default
        self.model = self.model.train()

    def on_validation_start(self) -> None:
        """Called at the beginning of validation."""
        # NOTE monkey patch result printing of the trainer
        try:
            trainer = self.trainer
        except RuntimeError:
            trainer = None
        if trainer is None:
            return

        trainer._evaluation_loop._print_results = lambda *args, **kwargs: None

    def on_test_start(self) -> None:
        """Called at the beginning of testing."""
        self.on_validation_start()

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """Configures the optizmizer for fine-tuning. This method is ignored when using the CLI. When using Lightning IR
        programmatically, the optimizer must be set using :meth:`set_optimizer`.

        Returns:
            torch.optim.Optimizer: The optimizer set for the model.
        Raises:
            ValueError: If optimizer is not set. Call `set_optimizer`.
        """
        if self._optimizer is None:
            raise ValueError("Optimizer is not set. Call `set_optimizer`.")
        return self._optimizer

    def set_optimizer(
        self, optimizer: type[torch.optim.Optimizer], **optimizer_kwargs: dict[str, Any]
    ) -> "LightningIRModule":
        """Sets the optimizer for the model. Necessary for fine-tuning when not using the CLI.

        Args:
            optimizer (type[torch.optim.Optimizer]): Torch optimizer class.
            optimizer_kwargs (dict[str, Any]): Arguments to initialize the optimizer.
        Returns:
            LightningIRModule: Self with the optimizer set.
        """
        self._optimizer = optimizer(self.parameters(), **optimizer_kwargs)
        return self

    def score(self, queries: Sequence[str] | str, docs: Sequence[Sequence[str]] | Sequence[str]) -> LightningIROutput:
        """Computes relevance scores for queries and documents.

        Args:
            queries (Sequence[str] | str): Queries to score.
            docs (Sequence[Sequence[str]] | Sequence[str]): Documents to score.
        Returns:
            LightningIROutput: Model output containing the scores.
        """
        if isinstance(queries, str):
            queries = (queries,)
        if isinstance(docs[0], str):
            docs = (docs,)
        batch = RankBatch(queries, docs, None, None)
        with torch.no_grad():
            return self.forward(batch)

    def forward(self, batch: TrainBatch | RankBatch | SearchBatch) -> LightningIROutput:
        """Handles the forward pass of the model.

        Args:
            batch (TrainBatch | RankBatch | SearchBatch): Batch of training or ranking data.
        Returns:
            LightningIROutput: Model output.
        Raises:
            NotImplementedError: Must be implemented by derived class.
        """
        raise NotImplementedError

    def prepare_input(
        self, queries: Sequence[str] | None, docs: Sequence[str] | None, num_docs: Sequence[int] | int | None
    ) -> dict[str, BatchEncoding]:
        """Tokenizes queries and documents and returns the tokenized BatchEncoding_.

        .. _BatchEncoding: https://huggingface.co/transformers/main_classes/tokenizer#transformers.BatchEncoding

        Args:
            queries (Sequence[str] | None): Queries to tokenize.
            docs (Sequence[str] | None): Documents to tokenize.
            num_docs (Sequence[int] | int | None): Number of documents per query, if None num_docs is inferred by
                `len(docs) // len(queries)`. Defaults to None.
        Returns:
            dict[str, BatchEncoding]: Tokenized queries and documents, format depends on the tokenizer.
        """
        encodings = self.tokenizer.tokenize(
            queries, docs, return_tensors="pt", padding=True, truncation=True, num_docs=num_docs
        )
        for key in encodings:
            encodings[key] = encodings[key].to(self.device)
        return encodings

    def _compute_losses(self, batch: TrainBatch, output: LightningIROutput) -> list[torch.Tensor]:
        """Computes the losses for a training batch."""
        raise NotImplementedError

    def training_step(self, batch: TrainBatch, batch_idx: int) -> dict[str, Any]:
        """Handles the training step for the model.

        Args:
            batch (TrainBatch): Batch of training data.
            batch_idx (int): Index of the batch.
        Returns:
            torch.Tensor: Sum of the losses weighted by the loss weights.
        Raises:
            ValueError: If no loss functions are set.
        """
        if self.loss_functions is None:
            raise ValueError("Loss functions are not set")
        output = self.forward(batch)
        losses = self._compute_losses(batch, output)
        total_loss = torch.tensor(0)
        assert len(losses) == len(self.loss_functions)
        for (loss_function, loss_weight), loss in zip(self.loss_functions, losses):
            self.log(loss_function.__class__.__name__, loss)
            total_loss = total_loss + loss * loss_weight
        self.log("loss", total_loss, prog_bar=True)
        return {"loss": total_loss, "output": output}

    def validation_step(
        self, batch: TrainBatch | RankBatch | SearchBatch, batch_idx: int, dataloader_idx: int = 0
    ) -> LightningIROutput:
        """Handles the validation step for the model.

        Args:
            batch (TrainBatch | RankBatch | SearchBatch): Batch of validation or testing data.
            batch_idx (int): Index of the batch.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        Returns:
            LightningIROutput: Model output.
        """
        output = self.forward(batch)

        if self.evaluation_metrics is None:
            return output

        dataset = self.get_dataset(dataloader_idx)
        dataset_id = str(dataloader_idx) if dataset is None else self.get_dataset_id(dataset)
        metrics = self.validate(output, batch)
        for key, value in metrics.items():
            key = f"{dataset_id}/{key}"
            self.log(key, value, batch_size=len(batch.queries))
        return output

    def test_step(
        self,
        batch: TrainBatch | RankBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> LightningIROutput:
        """Handles the testing step for the model. Passes the batch to the validation step.

        Args:
            batch (TrainBatch | RankBatch): Batch of testing data.
            batch_idx (int): Index of the batch.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        Returns:
            LightningIROutput: Model output.
        """
        return self.validation_step(batch, batch_idx, dataloader_idx)

    def get_dataset(self, dataloader_idx: int) -> IRDataset | None:
        """Gets the dataset instance from the dataloader index. Returns None if no dataset is found.

        Args:
            dataloader_idx (int): Index of the dataloader.
        Returns:
            IRDataset | None: Inference dataset or None if no dataset is found.
        """
        try:
            trainer = self.trainer
        except RuntimeError:
            trainer = None
        if trainer is None:
            return None
        STAGE_TO_DATALOADER = {
            RunningStage.VALIDATING: "val_dataloaders",
            RunningStage.TESTING: "test_dataloaders",
            RunningStage.PREDICTING: "predict_dataloaders",
            RunningStage.SANITY_CHECKING: "val_dataloaders",
        }
        if trainer.state.stage is None:
            return None
        dataloaders = getattr(trainer, STAGE_TO_DATALOADER[trainer.state.stage], None)
        if dataloaders is None:
            return None
        if isinstance(dataloaders, torch.utils.data.DataLoader):
            dataloaders = [dataloaders]
        return dataloaders[dataloader_idx].dataset

    def get_dataset_id(self, dataset: IRDataset) -> str:
        """Gets the dataset id from the dataloader index for logging.

        .. _ir-datasets: https://ir-datasets.com/

        Args:
            dataset (IRDataset): Dataset instance.
        Returns:
            str: Path to run file, ir-datasets_ dataset id, or dataloader index.
        """
        if isinstance(dataset, RunDataset) and dataset.run_path is not None:
            dataset_id = dataset.run_path.name
        else:
            dataset_id = dataset.dataset_id
        return dataset_id

    def validate(
        self,
        output: LightningIROutput,
        batch: TrainBatch | RankBatch | SearchBatch,
    ) -> dict[str, float]:
        """Validates the model output with the evaluation metrics and loss functions.

        Args:
            output (LightningIROutput): Model output.
            batch (TrainBatch | RankBatch | SearchBatch): Batch of validation or testing data.
        Returns:
            dict[str, float]: Dictionary of evaluation metrics.
        """
        metrics: dict[str, float] = {}
        if self.evaluation_metrics is None or output.scores is None:
            return metrics
        metrics.update(self.validate_metrics(output, batch))
        metrics.update(self.validate_loss(output, batch))
        return metrics

    def validate_metrics(
        self,
        output: LightningIROutput,
        batch: TrainBatch | RankBatch | SearchBatch,
    ) -> dict[str, float]:
        """Validates the model output with the evaluation metrics.

        Args:
            output (LightningIROutput): Model output.
            batch (TrainBatch | RankBatch | SearchBatch): Batch of validation or testing data.
        Returns:
            dict[str, float]: Dictionary of evaluation metrics.
        Raises:
            ValueError: If query_ids or doc_ids are not set in the batch.
        """
        metrics: dict[str, float] = {}
        qrels = batch.qrels
        if self.evaluation_metrics is None or qrels is None:
            return metrics
        query_ids = batch.query_ids
        doc_ids = batch.doc_ids
        if query_ids is None:
            raise ValueError("query_ids must be set")
        if doc_ids is None:
            raise ValueError("doc_ids must be set")
        evaluation_metrics = [metric for metric in self.evaluation_metrics if metric != "loss"]
        ir_measures_qrels = create_qrels_from_dicts(qrels)
        if evaluation_metrics and qrels is not None and output.scores is not None:
            run = create_run_from_scores(query_ids, doc_ids, output.scores)
            metrics.update(evaluate_run(run, ir_measures_qrels, evaluation_metrics))
        return metrics

    def validate_loss(
        self,
        output: LightningIROutput,
        batch: TrainBatch | RankBatch | SearchBatch,
    ) -> dict[str, float]:
        """Validates the model output with the loss functions.

        Args:
            output (LightningIROutput): Model output.
            batch (TrainBatch | RankBatch | SearchBatch): Batch of validation or testing data.
        Returns:
            dict[str, float]: Dictionary of evaluation metrics.
        """
        metrics: dict[str, float] = {}
        query_ids = batch.query_ids
        if query_ids is None:
            raise ValueError("query_ids must be set")
        if (
            self.evaluation_metrics is None
            or "loss" not in self.evaluation_metrics
            or getattr(batch, "targets", None) is None
            or self.loss_functions is None
            or output.scores is None
        ):
            return metrics
        output.scores = output.scores.view(len(query_ids), -1)
        for loss_function, _ in self.loss_functions:
            # NOTE skip in-batch losses because they can use a lot of memory
            if isinstance(loss_function, InBatchLossFunction):
                continue
            metrics[f"validation-{loss_function.__class__.__name__}"] = loss_function.compute_loss(output, batch)
        return metrics

    def on_validation_end(self) -> None:
        """Prints the validation results for each dataloader."""
        trainer = self.trainer
        if not (trainer.is_global_zero and trainer._evaluation_loop.verbose):
            return
        results = trainer.callback_metrics

        data = []
        for key, value in {**results, **self._additional_log_metrics}.items():
            if "dataloader_idx" in key:
                key = "/".join(key.split("/")[:-1])
            *dataset_parts, metric = key.split("/")
            if metric.startswith("validation-"):
                metric = metric[len("validation-") :]
            dataset = "/".join(dataset_parts)
            if isinstance(value, torch.Tensor):
                value = value.item()
            data.append({"dataset": dataset, "metric": metric, "value": value})
        if not data:
            return
        df = pd.DataFrame(data)
        df = df.pivot(index="dataset", columns="metric", values="value")
        df.columns.name = None

        # bring into correct order when skipping inference datasets
        datamodule = getattr(self.trainer, "datamodule", None)
        if datamodule is not None and hasattr(datamodule, "inference_datasets"):
            inference_datasets = datamodule.inference_datasets
            if len(inference_datasets) != df.shape[0]:
                raise ValueError(
                    "Number of inference datasets does not match number of dataloaders. "
                    "Check if the dataloaders are correctly configured."
                )
            dataset_ids = [self.get_dataset_id(dataset) for dataset in inference_datasets]
            df = df.reindex(dataset_ids)

        trainer.print(df)

    def on_test_end(self) -> None:
        """Prints the accumulated metrics for each dataloader."""
        self.on_validation_end()

    def save_pretrained(self, save_path: str | Path) -> None:
        """Saves the model and tokenizer to the save path.

        Args:
            save_path (str | Path): Path to save the model and tokenizer.
        """
        self.model.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)

    def on_save_checkpoint(self, checkpoint: dict[str, Any]) -> None:
        """Saves the model and tokenizer to the trainer's log directory."""
        if self.trainer is not None and self.trainer.log_dir is not None:
            if self.trainer.global_rank != 0:
                return
            _step = self.trainer.global_step
            self.config.save_step = _step
            log_dir = Path(self.trainer.log_dir)
            save_path = log_dir / "huggingface_checkpoint"
            self.save_pretrained(save_path)



from .ir_datasets_utils import ScoredDocTuples, register_new_dataset


def register_colbert_docpairs():
    dlc_contents = {
        "url": "https://huggingface.co/colbert-ir/colbertv2.0_msmarco_64way/resolve/main/examples.json?download=true",
        "expected_md5": "8be0c71e330ac54dcd77fba058d291c7",
        "cache_path": "msmarco-passage/train/colbert_64way.json",
    }
    register_new_dataset(
        "msmarco-passage/train/colbert-docpairs",
        docs="msmarco-passage",
        queries="msmarco-passage/train",
        qrels="msmarco-passage/train",
        docpairs=dlc_contents,
        DocpairsType=ScoredDocTuples,
    )



"""Base module for cross-encoder models.

This module provides the main classes and functions for cross-encoder models, including configurations, models,
modules, and tokenizers."""

from .cross_encoder_config import CrossEncoderConfig
from .cross_encoder_model import CrossEncoderModel, CrossEncoderOutput
from .cross_encoder_module import CrossEncoderModule
from .cross_encoder_tokenizer import CrossEncoderTokenizer

__all__ = [
    "CrossEncoderConfig",
    "CrossEncoderModel",
    "CrossEncoderModule",
    "CrossEncoderOutput",
    "CrossEncoderTokenizer",
]



"""
Module containing Lightning IR callbacks.

This module provides PyTorch Lightning compatible callbacks for various stages and tasks.
"""

from .callbacks import IndexCallback, RankCallback, RegisterLocalDatasetCallback, ReRankCallback, SearchCallback

__all__ = [
    "IndexCallback",
    "RankCallback",
    "RegisterLocalDatasetCallback",
    "ReRankCallback",
    "SearchCallback",
]



"""FAISS Indexer for Lightning IR Framework"""

import warnings
from pathlib import Path

import torch

from ...bi_encoder import BiEncoderModule, BiEncoderOutput
from ...data import IndexBatch
from ...models import ColConfig, DprConfig
from ..base import IndexConfig, Indexer


class FaissIndexer(Indexer):
    """Base class for FAISS indexers in the Lightning IR framework."""

    INDEX_FACTORY: str

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the FaissIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (FaissIndexConfig): Configuration for the FAISS index.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        Raises:
            ValueError: If the similarity function is not supported.
        """
        super().__init__(index_dir, index_config, module, verbose)
        import faiss

        similarity_function = self.module.config.similarity_function
        if similarity_function in ("cosine", "dot"):
            self.metric_type = faiss.METRIC_INNER_PRODUCT
        else:
            raise ValueError(f"similarity_function {similarity_function} unknown")

        index_factory = self.INDEX_FACTORY.format(**index_config.to_dict())
        if similarity_function == "cosine":
            index_factory = "L2norm," + index_factory
        self.index = faiss.index_factory(self.module.config.embedding_dim, index_factory, self.metric_type)

        self.set_verbosity()

        if torch.cuda.is_available():
            self.to_gpu()

    def to_gpu(self) -> None:
        """Move the FAISS index to GPU."""
        pass

    def to_cpu(self) -> None:
        """Move the FAISS index to CPU."""
        pass

    def set_verbosity(self, verbose: bool | None = None) -> None:
        """set the verbosity of the FAISS index.

        Args:
            verbose (bool | None): Whether to enable verbose output. If None, uses the index's current verbosity
                setting. Defaults to None.
        """
        self.index.verbose = self.verbose if verbose is None else verbose

    def process_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        """Process embeddings before adding them to the FAISS index.

        Args:
            embeddings (torch.Tensor): The embeddings to process.
        Returns:
            torch.Tensor: The processed embeddings.
        """
        return embeddings

    def save(self) -> None:
        """Save the FAISS index to disk.

        Raises:
            ValueError: If the number of embeddings does not match the index's total number of entries.
        """
        super().save()
        import faiss

        if self.num_embeddings != self.index.ntotal:
            raise ValueError("number of embeddings does not match index.ntotal")
        if torch.cuda.is_available() and hasattr(faiss, "index_gpu_to_cpu"):
            self.index = faiss.index_gpu_to_cpu(self.index)

        faiss.write_index(self.index, str(self.index_dir / "index.faiss"))

    def add(self, index_batch: IndexBatch, output: BiEncoderOutput) -> None:
        """Add embeddings to the FAISS index.

        Args:
            index_batch (IndexBatch): The batch containing document indices and embeddings.
            output (BiEncoderOutput): The output from the bi-encoder module containing document embeddings.
        Raises:
            ValueError: If the document embeddings are not present in the output.
        """
        doc_embeddings = output.doc_embeddings
        if doc_embeddings is None:
            raise ValueError("Expected doc_embeddings in BiEncoderOutput")
        if doc_embeddings.scoring_mask is None:
            doc_lengths = torch.ones(
                doc_embeddings.embeddings.shape[0], device=doc_embeddings.device, dtype=torch.int32
            )
            embeddings = doc_embeddings.embeddings[:, 0]
        else:
            doc_lengths = doc_embeddings.scoring_mask.sum(dim=1)
            embeddings = doc_embeddings.embeddings[doc_embeddings.scoring_mask]
        doc_ids = index_batch.doc_ids
        embeddings = self.process_embeddings(embeddings)

        if embeddings.shape[0]:
            self.index.add(embeddings.float().cpu())

        self.num_embeddings += embeddings.shape[0]
        self.num_docs += len(doc_ids)

        self.doc_lengths.extend(doc_lengths.int().cpu().tolist())
        self.doc_ids.extend(doc_ids)


class FaissFlatIndexer(FaissIndexer):
    """FAISS Flat Indexer for exact nearest neighbor search using FAISS."""

    INDEX_FACTORY = "Flat"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissFlatIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the FaissFlatIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (FaissFlatIndexConfig): Configuration for the FAISS flat index.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        """
        super().__init__(index_dir, index_config, module, verbose)
        self.index_config: FaissFlatIndexConfig

    def to_gpu(self) -> None:
        """Move the FAISS flat index to GPU."""
        pass

    def to_cpu(self) -> None:
        """Move the FAISS flat index to CPU."""
        pass


class _FaissTrainIndexer(FaissIndexer):
    """Base class for FAISS indexers that require training on embeddings before indexing."""

    INDEX_FACTORY = ""  # class only acts as mixin

    def __init__(
        self,
        index_dir: Path,
        index_config: "_FaissTrainIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the _FaissTrainIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (_FaissTrainIndexConfig): Configuration for the FAISS index that requires training.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        Raises:
            ValueError: If num_train_embeddings is not set in the index configuration.
        """
        super().__init__(index_dir, index_config, module, verbose)
        if index_config.num_train_embeddings is None:
            raise ValueError("num_train_embeddings must be set")
        self.num_train_embeddings = index_config.num_train_embeddings

        self._train_embeddings: torch.Tensor | None = torch.full(
            (self.num_train_embeddings, self.module.config.embedding_dim), torch.nan, dtype=torch.float32
        )

    def process_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        """Process embeddings before adding them to the FAISS index.

        Args:
            embeddings (torch.Tensor): The embeddings to process.
        Returns:
            torch.Tensor: The processed embeddings.
        """
        embeddings = self._grab_train_embeddings(embeddings)
        self._train()
        return embeddings

    def _grab_train_embeddings(self, embeddings: torch.Tensor) -> torch.Tensor:
        if self._train_embeddings is not None:
            # save training embeddings until num_train_embeddings is reached
            # if num_train_embeddings overflows, save the remaining embeddings
            start = self.num_embeddings
            end = start + embeddings.shape[0]
            end = min(self.num_train_embeddings, start + embeddings.shape[0])
            length = end - start
            self._train_embeddings[start:end] = embeddings[:length]
            self.num_embeddings += length
            embeddings = embeddings[length:]
        return embeddings

    def _train(self, force: bool = False):
        if self._train_embeddings is None:
            return
        if not force and self.num_embeddings < self.num_train_embeddings:
            return
        if torch.isnan(self._train_embeddings).any():
            warnings.warn(
                "Corpus contains less tokens/documents than num_train_embeddings. Removing NaN embeddings.",
                stacklevel=2,
            )
            self._train_embeddings = self._train_embeddings[~torch.isnan(self._train_embeddings).any(dim=1)]
        self.index.train(self._train_embeddings)
        if torch.cuda.is_available():
            self.to_cpu()
        self.index.add(self._train_embeddings)
        self._train_embeddings = None
        self.set_verbosity(False)

    def save(self) -> None:
        if not self.index.is_trained:
            self._train(force=True)
        return super().save()


class FaissIVFIndexer(_FaissTrainIndexer):
    """FAISS IVF Indexer for approximate nearest neighbor search using FAISS with Inverted File System (IVF)."""

    INDEX_FACTORY = "IVF{num_centroids},Flat"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIVFIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the FaissIVFIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (FaissIVFIndexConfig): Configuration for the FAISS IVF index.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        """
        # default faiss values
        # https://github.com/facebookresearch/faiss/blob/dafdff110489db7587b169a0afee8470f220d295/faiss/Clustering.h#L43
        max_points_per_centroid = 256
        index_config.num_train_embeddings = (
            index_config.num_train_embeddings or index_config.num_centroids * max_points_per_centroid
        )
        super().__init__(index_dir, index_config, module, verbose)

        import faiss

        ivf_index = faiss.extract_index_ivf(self.index)
        if hasattr(ivf_index, "quantizer"):
            quantizer = ivf_index.quantizer
            if hasattr(faiss.downcast_index(quantizer), "hnsw"):
                downcasted_quantizer = faiss.downcast_index(quantizer)
                downcasted_quantizer.hnsw.efConstruction = index_config.ef_construction

    def to_gpu(self) -> None:
        """Move the FAISS IVF index to GPU."""
        import faiss

        # clustering_index overrides the index used during clustering but leaves the quantizer on the gpu
        # https://faiss.ai/cpp_api/namespace/namespacefaiss_1_1gpu.html
        if faiss.get_num_gpus() == 0:
            return
        clustering_index = faiss.index_cpu_to_all_gpus(
            faiss.IndexFlat(self.module.config.embedding_dim, self.metric_type)
        )
        clustering_index.verbose = self.verbose
        index_ivf = faiss.extract_index_ivf(self.index)
        index_ivf.clustering_index = clustering_index

    def to_cpu(self) -> None:
        """Move the FAISS IVF index to CPU."""
        import faiss

        if faiss.get_num_gpus() == 0:
            return
        self.index = faiss.index_gpu_to_cpu(self.index)

        # https://gist.github.com/mdouze/334ad6a979ac3637f6d95e9091356d3e
        # move index to cpu but leave quantizer on gpu
        index_ivf = faiss.extract_index_ivf(self.index)
        quantizer = index_ivf.quantizer
        gpu_quantizer = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), 0, quantizer)
        index_ivf.quantizer = gpu_quantizer

    def set_verbosity(self, verbose: bool | None = None) -> None:
        """set the verbosity of the FAISS IVF index.

        Args:
            verbose (bool | None): Whether to enable verbose output. Defaults to None.
        """
        import faiss

        verbose = verbose if verbose is not None else self.verbose
        index = faiss.extract_index_ivf(self.index)
        for elem in (index, index.quantizer):
            elem.verbose = verbose


class FaissPQIndexer(_FaissTrainIndexer):
    """FAISS PQ Indexer for approximate nearest neighbor search using Product Quantization (PQ)."""

    INDEX_FACTORY = "OPQ{num_subquantizers},PQ{num_subquantizers}x{n_bits}"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissPQIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the FaissPQIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (FaissPQIndexConfig): Configuration for the FAISS PQ index.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        """
        super().__init__(index_dir, index_config, module, verbose)
        self.index_config: FaissPQIndexConfig

    def to_gpu(self) -> None:
        """Move the FAISS PQ index to GPU."""
        pass

    def to_cpu(self) -> None:
        """Move the FAISS PQ index to CPU."""
        pass


class FaissIVFPQIndexer(FaissIVFIndexer):
    """FAISS IVFPQ Indexer for approximate nearest neighbor search using Inverted File System (IVF) with Product
    Quantization (PQ)."""

    INDEX_FACTORY = "OPQ{num_subquantizers},IVF{num_centroids}_HNSW32,PQ{num_subquantizers}x{n_bits}"

    def __init__(
        self,
        index_dir: Path,
        index_config: "FaissIVFPQIndexConfig",
        module: BiEncoderModule,
        verbose: bool = False,
    ) -> None:
        """Initialize the FaissIVFPQIndexer.

        Args:
            index_dir (Path): Directory where the index will be stored.
            index_config (FaissIVFPQIndexConfig): Configuration for the FAISS IVFPQ index.
            module (BiEncoderModule): The BiEncoderModule to use for indexing.
            verbose (bool): Whether to enable verbose output. Defaults to False.
        """
        import faiss

        super().__init__(index_dir, index_config, module, verbose)
        self.index_config: FaissIVFPQIndexConfig

        index_ivf = faiss.extract_index_ivf(self.index)
        index_ivf.make_direct_map()

    def set_verbosity(self, verbose: bool | None = None) -> None:
        """set the verbosity of the FAISS IVFPQ index.

        Args:
            verbose (bool | None): Whether to enable verbose output. Defaults to None.
        """
        super().set_verbosity(verbose)
        import faiss

        verbose = verbose if verbose is not None else self.verbose
        index_ivf_pq = faiss.downcast_index(self.index.index)
        for elem in (
            index_ivf_pq.pq,
            index_ivf_pq.quantizer,
        ):
            elem.verbose = verbose


class FaissIndexConfig(IndexConfig):
    """Configuration class for FAISS indexers in the Lightning IR framework."""

    SUPPORTED_MODELS = {ColConfig.model_type, DprConfig.model_type}
    indexer_class: type[Indexer] = FaissIndexer


class FaissFlatIndexConfig(FaissIndexConfig):
    """Configuration class for FAISS flat indexers in the Lightning IR framework."""

    indexer_class = FaissFlatIndexer


class _FaissTrainIndexConfig(FaissIndexConfig):
    """Base configuration class for FAISS indexers that require training on embeddings before indexing."""

    indexer_class = _FaissTrainIndexer

    def __init__(self, num_train_embeddings: int | None = None) -> None:
        """Initialize the _FaissTrainIndexConfig.

        Args:
            num_train_embeddings (int | None): Number of embeddings to use for training the index. If None, it will
                be set later. Defaults to None.
        """
        super().__init__()
        self.num_train_embeddings = num_train_embeddings


class FaissIVFIndexConfig(_FaissTrainIndexConfig):
    """Configuration class for FAISS IVF indexers in the Lightning IR framework."""

    indexer_class = FaissIVFIndexer

    def __init__(
        self,
        num_train_embeddings: int | None = None,
        num_centroids: int = 262144,
        ef_construction: int = 40,
    ) -> None:
        """Initialize the FaissIVFIndexConfig.

        Args:
            num_train_embeddings (int | None): Number of embeddings to use for training the index. If None, it will be
                set later. Defaults to None.
            num_centroids (int): Number of centroids for the IVF index. Defaults to 262144.
            ef_construction (int): The size of the dynamic list used during construction. Defaults to 40.
        """
        super().__init__(num_train_embeddings)
        self.num_centroids = num_centroids
        self.ef_construction = ef_construction


class FaissPQIndexConfig(_FaissTrainIndexConfig):
    """Configuration class for FAISS PQ indexers in the Lightning IR framework."""

    indexer_class = FaissPQIndexer

    def __init__(self, num_train_embeddings: int | None = None, num_subquantizers: int = 16, n_bits: int = 8) -> None:
        """Initialize the FaissPQIndexConfig.

        Args:
            num_train_embeddings (int | None): Number of embeddings to use for training the index. If None, it will
                be set later. Defaults to None.
            num_subquantizers (int): Number of subquantizers for the PQ index. Defaults to 16.
            n_bits (int): Number of bits for the PQ index. Defaults to 8.
        """
        super().__init__(num_train_embeddings)
        self.num_subquantizers = num_subquantizers
        self.n_bits = n_bits


class FaissIVFPQIndexConfig(FaissIVFIndexConfig):
    """Configuration class for FAISS IVFPQ indexers in the Lightning IR framework."""

    indexer_class = FaissIVFPQIndexer

    def __init__(
        self,
        num_train_embeddings: int | None = None,
        num_centroids: int = 262144,
        ef_construction: int = 40,
        num_subquantizers: int = 16,
        n_bits: int = 8,
    ) -> None:
        """Initialize the FaissIVFPQIndexConfig.

        Args:
            num_train_embeddings (int | None): Number of embeddings to use for training the index. If None, it will
                be set later. Defaults to None.
            num_centroids (int): Number of centroids for the IVF index. Defaults to 262144.
            ef_construction (int): The size of the dynamic list used during construction. Defaults to 40.
            num_subquantizers (int): Number of subquantizers for the PQ index. Defaults to 16.
            n_bits (int): Number of bits for the PQ index. Defaults to 8.
        """
        super().__init__(num_train_embeddings, num_centroids, ef_construction)
        self.num_subquantizers = num_subquantizers
        self.n_bits = n_bits



"""
Pairwise loss functions for the Lightning IR framework.

This module contains loss functions that operate on pairs of items,
comparing positive and negative examples.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import torch

from .base import PairwiseLossFunction

if TYPE_CHECKING:
    from ..base import LightningIROutput
    from ..data import TrainBatch


class MarginMSE(PairwiseLossFunction):
    """Mean Squared Error loss with a margin for pairwise ranking tasks.
    Originally proposed in: `Improving Efficient Neural Ranking Models with Cross-Architecture Knowledge Distillation \
    <https://arxiv.org/abs/2010.02666>`_
    """

    def __init__(self, margin: float | Literal["scores"] = 1.0):
        """Initialize the MarginMSE loss function.

        Args:
            margin (float | Literal["scores"]): The margin value for the loss.
        """
        self.margin = margin

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the MarginMSE loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        Raises:
            ValueError: If the margin type is invalid.
        """
        scores = self.process_scores(output)
        targets = self.process_targets(scores, batch)
        query_idcs, pos_idcs, neg_idcs = self.get_pairwise_idcs(targets)
        pos = scores[query_idcs, pos_idcs]
        neg = scores[query_idcs, neg_idcs]
        margin = pos - neg
        if isinstance(self.margin, float):
            target_margin = torch.tensor(self.margin, device=scores.device)
        elif self.margin == "scores":
            target_margin = targets[query_idcs, pos_idcs] - targets[query_idcs, neg_idcs]
        else:
            raise ValueError("invalid margin type")
        loss = torch.nn.functional.mse_loss(margin, target_margin)
        return loss


class ConstantMarginMSE(MarginMSE):
    """Constant Margin MSE loss for pairwise ranking tasks with a fixed margin."""

    def __init__(self, margin: float = 1.0):
        """Initialize the ConstantMarginMSE loss function.

        Args:
            margin (float): The fixed margin value for the loss.
        """
        super().__init__(margin)


class SupervisedMarginMSE(MarginMSE):
    """Supervised Margin MSE loss for pairwise ranking tasks with a dynamic margin."""

    def __init__(self):
        """Initialize the SupervisedMarginMSE loss function."""
        super().__init__("scores")


class RankNet(PairwiseLossFunction):
    """RankNet loss function for pairwise ranking tasks.
    Originally proposed in: `Learning to Rank using Gradient Descent \
    <https://dl.acm.org/doi/10.1145/1102351.1102363>`_
    """

    def __init__(self, temperature: float = 1) -> None:
        super().__init__()
        """Initialize the RankNet loss function.
        Args:
            temperature (float): Temperature parameter for scaling the scores.
        """
        self.temperature = temperature

    def compute_loss(self, output: LightningIROutput, batch: TrainBatch) -> torch.Tensor:
        """Compute the RankNet loss.

        Args:
            output (LightningIROutput): The output from the model containing scores.
            batch (TrainBatch): The training batch containing targets.
        Returns:
            torch.Tensor: The computed loss.
        """
        scores = self.process_scores(output) * self.temperature
        targets = self.process_targets(scores, batch)
        query_idcs, pos_idcs, neg_idcs = self.get_pairwise_idcs(targets)
        pos = scores[query_idcs, pos_idcs]
        neg = scores[query_idcs, neg_idcs]
        margin = pos - neg
        loss = torch.nn.functional.binary_cross_entropy_with_logits(margin, torch.ones_like(margin))
        return loss



"""
Lightning IR data module.

This module provides classes for handling data in Lightning IR, including data modules, datasets, and data samples.
"""

from .data import DocSample, IndexBatch, QuerySample, RankBatch, RankSample, SearchBatch, TrainBatch
from .datamodule import LightningIRDataModule
from .dataset import DocDataset, IRDataset, QueryDataset, RunDataset, TupleDataset

__all__ = [
    "DocDataset",
    "DocSample",
    "IndexBatch",
    "IRDataset",
    "LightningIRDataModule",
    "QueryDataset",
    "QuerySample",
    "RankBatch",
    "RankSample",
    "RunDataset",
    "SearchBatch",
    "TrainBatch",
    "TupleDataset",
]



"""Configuration, model, and tokenizer for Col (Contextualized Late Interaction) type models. Originally proposed in
`ColBERT: Efficient and Effective Passage Search via Contextualized Late Interaction over BERT \
<https://dl.acm.org/doi/abs/10.1145/3397271.3401075>`_ as the ColBERT model. This implementation generalizes the model
to work with any transformer backbone model.
"""

from collections.abc import Sequence
from typing import Literal

import torch
from transformers import BatchEncoding

from ...bi_encoder import (
    BiEncoderEmbedding,
    BiEncoderOutput,
    BiEncoderTokenizer,
    MultiVectorBiEncoderConfig,
    MultiVectorBiEncoderModel,
)


class ColConfig(MultiVectorBiEncoderConfig):
    """Configuration class for a Col model."""

    model_type = "col"
    """Model type for a Col model."""

    def __init__(
        self,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        similarity_function: Literal["cosine", "dot"] = "dot",
        normalization_strategy: Literal["l2"] | None = None,
        add_marker_tokens: bool = False,
        query_mask_scoring_tokens: Sequence[str] | Literal["punctuation"] | None = None,
        doc_mask_scoring_tokens: Sequence[str] | Literal["punctuation"] | None = None,
        query_aggregation_function: Literal["sum", "mean", "max"] = "sum",
        doc_aggregation_function: Literal["sum", "mean", "max"] = "max",
        embedding_dim: int = 128,
        projection: Literal["linear", "linear_no_bias"] = "linear",
        query_expansion: bool = False,
        attend_to_query_expanded_tokens: bool = False,
        doc_expansion: bool = False,
        attend_to_doc_expanded_tokens: bool = False,
        k_train: int | None = None,
        **kwargs,
    ):
        """A Col model encodes queries and documents separately and computes a late interaction score between the query
        and document embeddings. The aggregation behavior of the late-interaction function can be parameterized with
        the `aggregation_function` arguments. The dimensionality of the token embeddings is down-projected using a
        linear layer. Queries and documents can optionally be expanded with mask tokens. Optionally, a set of tokens can
        be ignored during scoring.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            similarity_function (Literal["cosine", "dot"]): Similarity function to compute scores between query and
                document embeddings. Defaults to "dot".
            normalization_strategy (Literal['l2'] | None): Whether to normalize query and document embeddings.
                Defaults to None.
            add_marker_tokens (bool): Whether to add extra marker tokens [Q] / [D] to queries / documents.
                Defaults to False.
            query_mask_scoring_tokens (Sequence[str] | Literal["punctuation"] | None): Whether and which query tokens
                to ignore during scoring. Defaults to None.
            doc_mask_scoring_tokens (Sequence[str] | Literal["punctuation"] | None): Whether and which document tokens
                to ignore during scoring. Defaults to None.
            query_aggregation_function (Literal["sum", "mean", "max"]): How to aggregate
                similarity scores over query tokens. Defaults to "sum".
            doc_aggregation_function (Literal["sum", "mean", "max"]): How to aggregate
                similarity scores over document tokens. Defaults to "max".
            embedding_dim (int): The output embedding dimension. Defaults to 128.
            projection (Literal["linear", "linear_no_bias"]): Whether and how to project the output embeddings.
                Defaults to "linear". If set to "linear_no_bias", the projection layer will not have a bias term.
            query_expansion (bool): Whether to expand queries with mask tokens. Defaults to False.
            attend_to_query_expanded_tokens (bool): Whether to allow query tokens to attend to mask expanded query
                tokens. Defaults to False.
            doc_expansion (bool): Whether to expand documents with mask tokens. Defaults to False.
            attend_to_doc_expanded_tokens (bool): Whether to allow document tokens to attend to mask expanded document
                tokens. Defaults to False.
            k_train (int | None): Whether to use XTR_'s in-batch token retrieval during training and how many top-k
                document tokens to use. Defaults to 128.

        .. _XTR: \
https://proceedings.neurips.cc/paper_files/paper/2023/file/31d997278ee9069d6721bc194174bb4c-Paper-Conference.pdf
        """
        super().__init__(
            query_length=query_length,
            doc_length=doc_length,
            similarity_function=similarity_function,
            normalization_strategy=normalization_strategy,
            add_marker_tokens=add_marker_tokens,
            query_mask_scoring_tokens=query_mask_scoring_tokens,
            doc_mask_scoring_tokens=doc_mask_scoring_tokens,
            query_aggregation_function=query_aggregation_function,
            doc_aggregation_function=doc_aggregation_function,
            **kwargs,
        )
        self.embedding_dim = embedding_dim
        self.projection = projection
        self.query_expansion = query_expansion
        self.attend_to_query_expanded_tokens = attend_to_query_expanded_tokens
        self.doc_expansion = doc_expansion
        self.attend_to_doc_expanded_tokens = attend_to_doc_expanded_tokens
        self.k_train = k_train


class ColModel(MultiVectorBiEncoderModel):
    """Multi-vector late-interaction Col model. See :class:`.ColConfig` for configuration options."""

    config_class = ColConfig
    """Configuration class for the Col model."""

    def __init__(self, config: ColConfig, *args, **kwargs) -> None:
        """Initializes a Col model given a :class:`.ColConfig`.

        Args:
            config (ColConfig): Configuration for the Col model.
        Raises:
            ValueError: If the embedding dimension is not specified in the configuration.
        """
        super().__init__(config, *args, **kwargs)
        if config.embedding_dim is None:
            raise ValueError("Embedding dimension must be specified in the configuration.")
        self.projection = torch.nn.Linear(
            config.hidden_size, config.embedding_dim, bias="no_bias" not in config.projection
        )

    def score(
        self,
        output: BiEncoderOutput,
        num_docs: Sequence[int] | int | None = None,
    ) -> BiEncoderOutput:
        """Compute relevance scores between queries and documents.

        Args:
            output (BiEncoderOutput): Output containing embeddings and scoring mask.
            num_docs (Sequence[int] | int | None): Specifies how many documents are passed per query. If a sequence of
                integers, `len(num_doc)` should be equal to the number of queries and `sum(num_docs)` equal to the
                number of documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries. Defaults to None.
        Returns:
            BiEncoderOutput: Output containing relevance scores.
        """
        if self.training and self.config.k_train is not None:
            return self._score_xtr_in_batch(output, num_docs)

        return super().score(output, num_docs)

    def scoring_mask(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> torch.Tensor:
        """Computes a scoring mask for batched tokenized text sequences which is used in the scoring function to mask
        out vectors during scoring.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            torch.Tensor: Scoring mask.
        """
        input_ids = encoding["input_ids"]
        attention_mask = encoding["attention_mask"]
        scoring_mask = attention_mask
        expansion = getattr(self.config, f"{input_type}_expansion")
        if expansion or scoring_mask is None:
            scoring_mask = torch.ones_like(input_ids, dtype=torch.bool)
        scoring_mask = scoring_mask.bool()
        mask_scoring_input_ids = getattr(self, f"{input_type}_mask_scoring_input_ids")
        if mask_scoring_input_ids is not None:
            ignore_mask = input_ids[..., None].eq(mask_scoring_input_ids.to(input_ids.device)).any(-1)
            scoring_mask = scoring_mask & ~ignore_mask
        return scoring_mask

    def encode(self, encoding: BatchEncoding, input_type: Literal["query", "doc"]) -> BiEncoderEmbedding:
        """Encodes a batched tokenized text sequences and returns the embeddings and scoring mask.

        Args:
            encoding (BatchEncoding): Tokenizer encodings for the text sequence.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BiEncoderEmbedding: Embeddings and scoring mask.
        """
        embeddings = self._backbone_forward(**encoding).last_hidden_state
        embeddings = self.projection(embeddings)
        if self.config.normalization_strategy == "l2":
            embeddings = torch.nn.functional.normalize(embeddings, dim=-1)
        scoring_mask = self.scoring_mask(encoding, input_type)
        return BiEncoderEmbedding(embeddings, scoring_mask, encoding)

    def _score_xtr_in_batch(
        self, output: BiEncoderOutput, num_docs: Sequence[int] | int | None = None
    ) -> BiEncoderOutput:
        """XTR in-batch token retrieval scoring.

        Args:
            output (BiEncoderOutput): Output containing embeddings and scoring mask.
            num_docs (Sequence[int] | int | None): Specifies how many documents are passed per query. If a sequence of
                integers, `len(num_doc)` should be equal to the number of queries and `sum(num_docs)` equal to the
                number of documents, i.e., the sequence contains one value per query specifying the number of documents
                for that query. If an integer, assumes an equal number of documents per query. If None, tries to infer
                the number of documents by dividing the number of documents by the number of queries. Defaults to None.
        Returns:
            BiEncoderOutput: Output containing relevance scores.
        """
        query_embeddings = output.query_embeddings
        doc_embeddings = output.doc_embeddings
        if query_embeddings is None or doc_embeddings is None:
            raise ValueError("Both query and document embeddings must be provided for scoring.")
        similarities = self.compute_similarity(query_embeddings, doc_embeddings, num_docs)

        query_mask = query_embeddings.scoring_mask
        doc_mask = doc_embeddings.scoring_mask

        num_docs_t = self._parse_num_docs(
            query_embeddings.embeddings.shape[0],
            doc_embeddings.embeddings.shape[0],
            num_docs,
            query_embeddings.device,
        )

        query_mask_expanded = query_mask.repeat_interleave(num_docs_t, dim=0).unsqueeze(-1)
        doc_mask_expanded = doc_mask.unsqueeze(1)

        similarities = similarities.masked_fill(~doc_mask_expanded, float("-inf"))
        similarities = similarities.masked_fill(~query_mask_expanded, float("-inf"))

        batch_size = query_embeddings.embeddings.shape[0]
        q_len = query_embeddings.embeddings.shape[1]
        doc_len = doc_embeddings.embeddings.shape[1]
        max_docs = torch.max(num_docs_t)

        sim_list = similarities.split(num_docs_t.tolist(), dim=0)
        sim_padded = torch.nn.utils.rnn.pad_sequence(sim_list, batch_first=True, padding_value=float("-inf"))

        valid_mask = torch.arange(max_docs, device=num_docs_t.device).unsqueeze(0) < num_docs_t.unsqueeze(1)

        sim_flat = sim_padded.view(batch_size, -1)
        k_train = min(self.config.k_train, sim_flat.size(-1))
        minimum_values = torch.topk(sim_flat, k=k_train, dim=-1).values[:, -1].unsqueeze(-1)

        sim_padded = sim_padded.view(batch_size, -1)
        sim_padded = sim_padded.masked_fill(sim_padded < minimum_values, 0.0)
        sim_padded = sim_padded.view(batch_size, max_docs, q_len, doc_len)

        scores = sim_padded.max(dim=-1).values.sum(dim=-1)
        Z = (sim_padded.max(dim=-1).values > 0).sum(dim=-1).float()
        Z = Z.clamp(min=1.0)
        scores = scores / Z

        scores = scores[valid_mask]

        output.scores = scores
        output.similarity = sim_padded[valid_mask]

        return output


class ColTokenizer(BiEncoderTokenizer):
    """:class:`.LightningIRTokenizer` for Col models."""

    config_class = ColConfig
    """Configuration class for the tokenizer."""

    def __init__(
        self,
        *args,
        query_length: int | None = 32,
        doc_length: int | None = 512,
        add_marker_tokens: bool = False,
        query_expansion: bool = False,
        attend_to_query_expanded_tokens: bool = False,
        doc_expansion: bool = False,
        attend_to_doc_expanded_tokens: bool = False,
        **kwargs,
    ):
        """Initializes a Col model's tokenizer. Encodes queries and documents separately. Optionally adds marker tokens
        to encoded input sequences and expands queries and documents with mask tokens.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
            add_marker_tokens (bool): Whether to add extra marker tokens [Q] / [D] to queries / documents.
                Defaults to False.
            query_expansion (bool): Whether to expand queries with mask tokens. Defaults to False.
            attend_to_query_expanded_tokens (bool): Whether to allow query tokens to attend to mask expanded query
                tokens. Defaults to False.
            doc_expansion (bool): Whether to expand documents with mask tokens. Defaults to False.
            attend_to_doc_expanded_tokens (bool): Whether to allow document tokens to attend to mask expanded document
                tokens. Defaults to False.
        Raises:
            ValueError: If `add_marker_tokens` is True and a non-supported tokenizer is used.
        """
        super().__init__(
            *args,
            query_length=query_length,
            doc_length=doc_length,
            query_expansion=query_expansion,
            attend_to_query_expanded_tokens=attend_to_query_expanded_tokens,
            doc_expansion=doc_expansion,
            attend_to_doc_expanded_tokens=attend_to_doc_expanded_tokens,
            add_marker_tokens=add_marker_tokens,
            **kwargs,
        )
        self.query_expansion = query_expansion
        self.attend_to_query_expanded_tokens = attend_to_query_expanded_tokens
        self.doc_expansion = doc_expansion
        self.attend_to_doc_expanded_tokens = attend_to_doc_expanded_tokens

    def _expand(self, encoding: BatchEncoding, attend_to_expanded_tokens: bool) -> BatchEncoding:
        """Applies mask expansion to the input encoding."""
        input_ids = encoding["input_ids"]
        input_ids[input_ids == self.pad_token_id] = self.mask_token_id
        encoding["input_ids"] = input_ids
        if attend_to_expanded_tokens:
            encoding["attention_mask"].fill_(1)
        return encoding

    def tokenize_input_sequence(
        self, text: Sequence[str] | str, input_type: Literal["query", "doc"], *args, **kwargs
    ) -> BatchEncoding:
        """Tokenizes an input sequence. This method is used to tokenize both queries and documents.

        Args:
            text (Sequence[str] | str): Input text to tokenize.
            input_type (Literal["query", "doc"]): type of input, either "query" or "doc".
        Returns:
            BatchEncoding: Tokenized input sequences.
        """
        expansion = getattr(self, f"{input_type}_expansion")
        if expansion:
            kwargs["padding"] = "max_length"
        encoding = super().tokenize_input_sequence(text, input_type, *args, **kwargs)
        if expansion:
            encoding = self._expand(encoding, getattr(self, f"attend_to_{input_type}_expanded_tokens"))
        return encoding



"""
Tokenizer module for Lightning IR.

This module contains the main tokenizer class for the Lightning IR library.
"""

import json
from collections.abc import Sequence
from os import PathLike
from typing import Self

from transformers import TOKENIZER_MAPPING, BatchEncoding, PreTrainedTokenizerBase

from .class_factory import LightningIRTokenizerClassFactory
from .config import LightningIRConfig
from .external_model_hub import CHECKPOINT_MAPPING


class LightningIRTokenizer(PreTrainedTokenizerBase):
    """Base class for Lightning IR tokenizers. Derived classes implement the tokenize method for handling query
    and document tokenization. It acts as mixin for a transformers.PreTrainedTokenizer_ backbone tokenizer.

    .. _transformers.PreTrainedTokenizer: \
https://huggingface.co/transformers/main_classes/tokenizer.htmltransformers.PreTrainedTokenizer
    """

    config_class: type[LightningIRConfig] = LightningIRConfig
    """Configuration class for the tokenizer."""

    def __init__(self, *args, query_length: int | None = 32, doc_length: int | None = 512, **kwargs):
        """Initializes the tokenizer.

        Args:
            query_length (int | None): Maximum number of tokens per query. If None does not truncate. Defaults to 32.
            doc_length (int | None): Maximum number of tokens per document. If None does not truncate. Defaults to 512.
        """
        super().__init__(*args, query_length=query_length, doc_length=doc_length, **kwargs)
        self.query_length = query_length
        self.doc_length = doc_length

    def tokenize(
        self, queries: str | Sequence[str] | None = None, docs: str | Sequence[str] | None = None, **kwargs
    ) -> dict[str, BatchEncoding]:
        """Tokenizes queries and documents.

        Args:
            queries (str | Sequence[str] | None): Queries to tokenize. Defaults to None.
            docs (str | Sequence[str] | None): Documents to tokenize. Defaults to None.
        Returns:
            dict[str, BatchEncoding]: Dictionary containing tokenized queries and documents.
        Raises:
            NotImplementedError: Must be implemented by the derived class.
        """
        raise NotImplementedError

    @classmethod
    def from_pretrained(cls, model_name_or_path: str, *args, **kwargs) -> Self:
        """Loads a pretrained tokenizer. Wraps the transformers.PreTrainedTokenizer.from_pretrained_ method to return a
        derived LightningIRTokenizer class. See :class:`.LightningIRTokenizerClassFactory` for more details.

        .. _transformers.PreTrainedTokenizer.from_pretrained: \
https://huggingface.co/docs/transformers/main_classes/tokenizer.html#transformers.PreTrainedTokenizer.from_pretrained

        .. highlight:: python
        .. code-block:: python

            >>> Loading using model class and backbone checkpoint
            >>> type(BiEncoderTokenizer.from_pretrained("bert-base-uncased"))
            ...
            <class 'lightning_ir.base.class_factory.BiEncoderBertTokenizerFast'>
            >>> Loading using base class and backbone checkpoint
            >>> type(LightningIRTokenizer.from_pretrained("bert-base-uncased", config=BiEncoderConfig()))
            ...
            <class 'lightning_ir.base.class_factory.BiEncoderBertTokenizerFast'>

        Args:
            model_name_or_path (str): Name or path of the pretrained tokenizer.
        Returns:
            Self: A derived LightningIRTokenizer consisting of a backbone tokenizer and a LightningIRTokenizer mixin.
        Raises:
            ValueError: If called on the abstract class `LightningIRTokenizer` and no config is passed.
        """
        # provides AutoTokenizer.from_pretrained support
        config = kwargs.get("config", None)
        if cls is LightningIRTokenizer or all(issubclass(base, LightningIRTokenizer) for base in cls.__bases__):
            # no backbone models found, create derived lightning-ir tokenizer based on backbone model
            if config is not None:
                ConfigClass = config.__class__
            elif model_name_or_path in CHECKPOINT_MAPPING:
                _config = CHECKPOINT_MAPPING[model_name_or_path]
                ConfigClass = _config.__class__
                if config is None:
                    kwargs["config"] = _config
            elif cls is not LightningIRTokenizer and hasattr(cls, "config_class"):
                ConfigClass = cls.config_class
            else:
                ConfigClass = LightningIRTokenizerClassFactory.get_lightning_ir_config(model_name_or_path)
                if ConfigClass is None:
                    raise ValueError("Pass a config to `from_pretrained`.")
            ConfigClass = getattr(ConfigClass, "mixin_config", ConfigClass)
            backbone_config = LightningIRTokenizerClassFactory.get_backbone_config(model_name_or_path)
            BackboneTokenizers = TOKENIZER_MAPPING[type(backbone_config)]
            if kwargs.get("use_fast", True):
                BackboneTokenizer = BackboneTokenizers[1]
            else:
                BackboneTokenizer = BackboneTokenizers[0]
            cls = LightningIRTokenizerClassFactory(ConfigClass).from_backbone_class(BackboneTokenizer)
            return cls.from_pretrained(model_name_or_path, *args, **kwargs)
        config = kwargs.pop("config", None)
        if config is not None:
            kwargs.update(config.get_tokenizer_kwargs(cls))
        return super().from_pretrained(model_name_or_path, *args, **kwargs)

    def _save_pretrained(
        self,
        save_directory: str | PathLike,
        file_names: tuple[str],
        legacy_format: bool | None = None,
        filename_prefix: str | None = None,
    ) -> tuple[str]:
        # bit of a hack to change the tokenizer class in the stored tokenizer config to only contain the
        # lightning_ir tokenizer class (removing the backbone tokenizer class)
        save_files = super()._save_pretrained(save_directory, file_names, legacy_format, filename_prefix)
        config_file = save_files[0]
        with open(config_file) as file:
            tokenizer_config = json.load(file)

        tokenizer_class = None
        backbone_tokenizer_class = None
        for base in self.__class__.__bases__:
            if issubclass(base, LightningIRTokenizer):
                if tokenizer_class is not None:
                    raise ValueError("Multiple Lightning IR tokenizer classes found.")
                tokenizer_class = base.__name__
                continue
            if issubclass(base, PreTrainedTokenizerBase):
                backbone_tokenizer_class = base.__name__

        tokenizer_config["tokenizer_class"] = tokenizer_class
        tokenizer_config["backbone_tokenizer_class"] = backbone_tokenizer_class

        with open(config_file, "w") as file:
            out_str = json.dumps(tokenizer_config, indent=2, sort_keys=True, ensure_ascii=False) + "\n"
            file.write(out_str)
        return save_files
